/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_AUTHORIZE_HH
#define SENDS_AUTHORIZE_HH

#include "seda_stage.hh"
#include "buffer_pool.hh"


namespace sends {

    /**  This stage performs the user authentication and copies the user
      *  data to the transaction object. The transaction is then passed on 
      *  to the following stage.
      *  @brief User authorization SEDA stage.
      *  @author John Zweizig
      *  @version 1.0: Last modified January 23, 2008
      *  @ingroup sends_stage
      */
    class authorize : public seda_stage {
    public:
	/**  Construct an authorization stage and thread pool consisting
	  *  of nThread threads, with an input queue of nTrans slots and
	  *  a resource queue with nBuffer buffers.
	  *  @brief Contructor.
	  *  @param nThread Number of threads in the authorization thread pool.
	  *  @param nTrans Maximum number of transactions in th input queue.
	  *  @param nBuffer Number of buffers in the resource queue.
	  */
	authorize(int nThread, int nTrans, int nBuffer=128);

	/**  Destroy the authorization stage.
	  *  @brief Destructor.
	  */
	~authorize(void);

	/**  Return a modifiable reference to the resource buffer queue.
	  *  @brief get resource queue.
	  *  @return Reference to the resource queue.
	  */
	buffer_pool& ref_resource(void);

	/**  Set the parameters for this stage.
	  *  @brief Set stage parameters.
	  *  @param p parameter block.
	  */
	void set_parameters(const stage_pars& p);

	/**  Complete the connection handshake with the client and perform 
	  *  the user authentication processing. At present (2008.01.23) 
	  *  there is no user authentication, although the stage reads a
	  *  string that will be used for this purpose in the future.
	  *  @brief Perform authorization processing.
	  *  @param trans Transaction object to be processed by this stage.
	  */
	void filter(transaction* trans);

	/**  Return a constant pointer to a null-terminated character vector 
	  *  with the stage name ("authorization").
	  *  @brief Stage name.
	  *  @return Pointer to the stage name string.
	  */
	const char* stage_name(void) const;

    private:
	buffer_pool resource;
	double timeout;
    };

    inline buffer_pool&
    authorize::ref_resource(void) {
	return resource;
    }
}

#endif // !defined(SENDS_AUTHORIZE_HH)
