/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_AVAILABLE_HH
#define SENDS_AVAILABLE_HH

#include "epoch.hh"
#include "sortedvector.hh"
#include <string>
#include <iosfwd>

namespace sends {

    class chan_entry;
    class loc_list;

    /** the available class holds a vector of epochs and can perform logic 
     * (and or) on them.
     */
    class available {
    public:
	/**  Construct an empty availability list.
	  *  \brief default constructor
	  */
	available(void);

	/** Clear the list.
	  * \brief Clear list.
	 */
	void clear(void);

	/**  Coalesce adjacent or overlapping entries in an availability list.
	  *  \brief Coalesce adjacent entries.
	  */
	void coalesce(void);

	/**  Test for empty list.
	  *  \brief test empty list.
	  *  \return True if there are no epochs in the list.
	 */
	bool empty(void) const;

	/** Calculate intersection of two availability lists.
 	  * \brief Intersection with \a x.
 	  * \param x List with which to find intersection.
 	  * \return Reference to modified list. 
	  */
	available& operator&=(const available& x);

	/** Calculate intersection of list with an epoch.
 	  * \brief Intersection with epoch.
 	  * \param x Epoch with which to find intersection.
 	  * \return Reference to modified list. 
	  */
	available& operator&=(const epoch& x);

	/**  Calculate list containing combinned epochs in two lists.
	  *  \brief Combined list.
	  *  \param x List to combine with current instance.
	  *  \return Reference to modified list.
	  */ 
	available& operator|=(const available& x);

	/**  Reference a specified list element.
	  *  \brief Reference an epoch element.
	  *  \param i index into list of element to be referenced.
	  *  \retuen Reference to specified element.
	  */
	const epoch& operator[](size_t i) const;

	/**  Set the list from the source elements of a specified source type.
	  *  \brief Set list from channel source data.
	  *  \param cl     Constant reference to channel entry.
	  *  \param source Source frame type string.
	  */
	void set_list(const chan_entry& cl, const std::string& source);

	/**  Set the list from the available data in a specified location list.
	  *  \brief Set availability from frame location list.
	  *  \param sl Constant reference to the location list.
	  */
	void set_list(const loc_list& sl);

	/**  Number of epochs in the list.
	  *  \brief List size
	  *  \return Number of epoch entries in availability list.
	  */
	size_t size(void) const;

	/**  Sort the list entries. Since the list is supposed to be sorted, 
	  *  this should only be necessary if some non-trivial modification
	  *  is made.
	  *  \brief Sort the list.
	  */
	void sort(void);

	/**  write a list to the specified output stream.
	  *  \brief Write list to output steam.
	  *  \param src Header string for each element.
	  *  \param out Output stream reference.
	  */
	void src_list(const std::string& src, std::ostream& out) const;
    private:
	typedef sorted_vector<epoch> epik_vect;
	epik_vect _vect;
    };

    /** Output operator.
     */
    std::ostream& operator<<(std::ostream& out, const available& av);

    //==================================  Inline methods.
    inline bool
    available::empty(void) const {
	return _vect.empty();
    }

    inline const epoch& 
    available::operator[](size_t i) const {
	return _vect[i];
    }

    inline size_t
    available::size(void) const {
	return _vect.size();
    }

}
#endif  // !defined(SENDS_AVAILABLE_HH)
