/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_BUFFER_POOL_HH
#define SENDS_BUFFER_POOL_HH

//#include "tbb/concurrent_queue.h"
#include "mt_queue.hh"
#include "types.hh"

namespace sends {

    class buffer;

    /**  The buffer_pool class defines a pool  of preallocated fixed length 
      *  buffers. 
      *  @brief Preallocated buffer pool.
      *  @author John Zweizig
      *  @version 1.0; Last modified January 31, 2008
      *  @ingroup sends_base
      */
    class buffer_pool {
    public:
	/**  Free buffer queue.
	  */
	typedef thread::mt_queue<buffer*> buffer_queue;
	//typedef tbb::concurrent_queue<buffer*> buffer_queue;
    public:
	/**  Construct a pool with nbuf buffers of length lbuf.
	  *  @brief Pool constructor
	  *  @param nbuf Number of buffers to allocate.
	  *  @param lbuf Buffer length.
	  */
	buffer_pool(long nbuf, long lbuf);

	/**  Destroy a buffer pool.
	  *  @brief Destructor
	  */
	~buffer_pool(void);

	/**  Get the originally allocated buffer size.
	  *  @brief Buffer size.
	  *  @return Originally allocated buffer size.
	  */
	index_type buffer_size(void) const;

	/**  Get a free buffer from the pool. If no buffers are currently 
	  *  available, the get_buffer request will block until a buffer
	  *  is return to the pool.
	  *  @brief Get a buffer
	  *  @return Buffer pointer.
	  */
	buffer* get_buffer(void);

	/**  Get a free buffer from the pool. Return a null pointer if no 
	  *  buffers are immediately available.
	  *  @brief Get a buffer (nowait)
	  *  @return Buffer pointer or null.
	  */
	buffer* get_buffer_now(void);

	/**  Return a free buffer to the pool.
	  *  @brief Return a buffer.
	  *  @param buf Pointer to the buffer to be returned.
	  */
	void return_buffer(buffer* buf);

	/**  Get the number of buffers assigned to the pool.
	  *  @brief Buffer count.
	  *  @return Number of buffers allocated to the pool.
	  */
	index_type size(void) const;
    private:
	index_type   mBufSize;
	buffer_queue mQueue;
    };

    //==================================  Inline functions
    inline index_type
    buffer_pool::buffer_size(void) const {
	return mBufSize;
    }

    inline index_type
    buffer_pool::size(void) const {
	return mQueue.capacity();
    }
}

#endif
