/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_CHAN_TYPE_HH
#define SENDS_CHAN_TYPE_HH
#include <string>

namespace sends {

    /**  The chan_type class enumerates the possible channel types and 
      *  implements conversion between character and coded representations.
      *  \brief Enumerate and handle channel types.
      *  \author John Zweizig (john.zweizig@ligo.org).
      *  \version 1.0; Last modified May 13, 2010.
      */
    class chan_type {
    public:
	
	/** Enumerate channel types.
	  */
	enum chan_enum {
	    kUnknown,	    ///< Unknown channel type
	    kOnline,	    ///< Online channels
	    kRawData,	    ///< Channels from raw data frames (FrAdcData)
	    kReduced,	    ///< Processed data from reduced frames (FrProcData)
	    kSecondTrend,   ///< Second trends
	    kMinuteTrend,   ///< Minute trends
	    kTestPt,	    ///< Test Point
	    kStatic,	    ///< Static calibration data (FrStatData)
	    kSimData,       ///< Simulated data
	    kTypeCount      ///< Number of data types
	};

    public:
	/**  Create a default channel type object. The initial channel
	  *  type is kUnknown.
	  *  \brief Default constructor.
	  */
	chan_type(void);

	/**  Construct a chan_type obect with the specified enumerated 
	  *  channel type.
	  *  \brief Type constructor
	  *  \param c Enumerated channel type.
	  */
	chan_type(chan_enum c);

	/**  Construct a chan_type object and set the type corresponding 
	  *  to the argument string. A runtime_error exception is thrown 
	  *  if the string does not name a valid channel type.
	  *  \brief Construct type named by string.
	  *  \param type Channel type name.
	 */
	explicit chan_type(const std::string& type);

	/**  Get the enumerated type code.
	  *  \brief Type accessor
	  */
	int code(void) const;

	/**  Compare this channel type to the argument type.
	  *  \brief Compare types
	  *  \param c Channel type to be comared to this
	  *  \return true if this is less than the argument type in enum order.
	  */
	bool operator<(const chan_type& c) const;

	/**  Compare this channel type to the argument type.
	  *  \brief Compare types
	  *  \param c Channel type to be comared to this
	  *  \return true if this is equal to the argument type.
	  */
	bool operator==(const chan_type& c) const;

	/**  Compare this channel type to the argument type.
	  *  \brief Compare types
	  *  \param c Channel type to be comared to this
	  *  \return true if this is not equal to the argument type.
	  */
	bool operator!=(const chan_type& c) const;

	/**  Set the channel type to that named by a string. A runtime_error
	  *  exception is thrown if the name string is not a valid channel 
	  *  type.
	  *  \brief Set chan_type to a named channel type.
	  *  \param type Channel type name string.
	  */
	void set(const std::string& type);

	/**  Set the channel type to that named by a string. A runtime_error
	  *  exception is thrown if the name string is not a valid channel 
	  *  type.
	  *  \brief Set chan_type to a named channel type.
	  *  \param type Channel type name string.
	  */
	void set(const char* type);

	/**  The channel type name strin is returned. 
	  *  \brief String accessor
	  *  \return Constant character pointer to the channel type name string.
	  */
	const char* str(void) const;
	
    private:
	chan_enum mChanType;
    };

    //==================================  Inline methods
    inline
    chan_type::chan_type(void)
      : mChanType(kUnknown)
    {}

    inline
    chan_type::chan_type(chan_enum type)
      : mChanType(type)
    {}

    inline
    chan_type::chan_type(const std::string& s) {
	set(s);
    }

    inline int
    chan_type::code(void) const {
	return int(mChanType);
    }

    inline bool 
    chan_type::operator<(const chan_type& c) const {
	return int(mChanType) < int(c.mChanType);
    }

    inline bool 
    chan_type::operator==(const chan_type& c) const {
	return int(mChanType) == int(c.mChanType);
    }

    inline bool 
    chan_type::operator!=(const chan_type& c) const {
	return int(mChanType) != int(c.mChanType);
    }


} // namespace sends

#endif // !defined(SENDS_CHAN_TYPE_HH)
