/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_TEMPLATE_STAGE_HH
#define SENDS_TEMPLATE_STAGE_HH

#include "seda_stage.hh"
#include <vector>
#include <atomic>
#include <gmutex.hh>

namespace sends {


    /**  The concentrator class handles multiple transaction waits. A 
      *  wait_data() command is issued on each transaction entered into 
      *  the input queue. The transaction is passed to the output queue
      *  if data arrives or is release if the wait request times out. The
      *  concentrator stage fits between the authentication srtage and the
      *  command parser stage. The preliminary implementation has one thread 
      *  per waiting transaction. In the future, I intend to modify it to 
      *  concentrate multiple waits into a single thread using the select 
      *  mechanism.
      *  @brief SEDA wait concentrator stage class.
      *  @author J. Zweizig
      *  @version 1.0; Last modified December 28, 2007
      */
    class concentrator : public seda_stage {
    public:
	/**  Construct a concentrator object. The arguments indicate the 
	  *  number of threads in the pool and the number of input transactions
	  *  that may be queue simultaneously (note that the pipline gets 
	  *  indigestion only after more than (nTrans + nThread) trasactions
	  *  need to execute in this stage.
	  *  @brief Concentrator pool constructor.
	  *  @param nThread Number of threads to run simultaneously in the pool.
	  *  @param nTrans  Number of transactions that may be queued.
	  */
	concentrator(int nThread, int nTrans);

	/**  Concentrator destructor.
	  */
	virtual ~concentrator(void);

	/**  Add a transaction to the list of sockets waiting for input.
	  *  \brief Add transaction to wait list.
	  *  \param trans Pointer to transaction object.
	  *  \param timeout Time (in seconds) to wait before moving the 
	  *                 transaction to the error queue.
	  */
	void add_transaction(transaction* trans, double timeout);

	/**  The initial implementation of the concentrator filter waits on
	  *  a single input transaction.
	  *  @brief Transaction filter for this stage.
	  *  @param trans Pointer to the current transaction 
	  */
	void filter(transaction* trans);

	/**  Initialize signal handling for this thread. The SIGPIPE signal
	  *  should leak through the select...
	  *  @brief Thread initialization.
	  */
	void init_thread(void);

	/**  Set the parameters for this stage.
	  *  @brief Set stage parameters.
	  *  @param p parameter block.
	  */
	void set_parameters(const stage_pars& p);

	/**  Return a constant pointer to a null-terminated character vector 
	  *  containing the name of the stage.
	  *  @brief Stage name.
	  *  @return Constant pointer to the stage name string.
	  */
	const char* stage_name(void) const;

    private:
	void lock(void);
	void unlock(void);
	bool try_collect(void);
	void release_collect(void);

    private:
	struct hold {
	    hold(transaction* tx, const timespec& t);
	    timespec     timeout;
	    transaction* trans;
	    int          socket;
	    bool         collected;
	};

	typedef std::vector<hold> hold_vect;
	typedef hold_vect::iterator hold_iter;

    private:
	double        mTimeout;

	thread::mutex    gate;
	std::atomic_flag collector;
	hold_vect        trans_vect;
    };
}

#endif // !defined(SENDS_CONCENTRATOR_HH)
