/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_DATA_TYPE_HH
#define SENDS_DATA_TYPE_HH
#include <string>

namespace sends {
    /**  The data type class enumerates the data types and provides information
      *  about the data types (e.g. word length) and translation between 
      *  string and enumerated code.
      *  \brief Data type class.
      *  \author John Zweizig (john.zweizig@ligo.org)
      *  \version 1.0; Last  modified May 13, 2010
      */
    class data_type {
    public:

	/** Enumerate data types.
	  */
	enum data_enum {

	    /// Char (1-byte integers).
	    kByte1,

	    /// Short (2-byte integers).
	    kInt2,

	    /// Int (4-byte integers).
	    kInt4,

	    /// Long (8-byte integers).
	    kInt8,

	    /// Floats (4-byte float)
	    kReal4,

	    /// Doubles (8-byte float)
	    kReal8,

	    /// Float Complex (2x4-byte float)
	    kComplex8,

	    /// Unsigned integer
	    kUInt4,

	    /// Double Complex (2x8-byte float)
	    kComplex16     // Number of data types
	};

    public:
	/**  A data type object is defined with a default data type of 
	  *  kByte1.
	  *  \brief Default constructor.
	  */
	data_type(void);

	/**  A data type  object is constructed with the specified data
	  *  type value.
	  *  \brief Data type constructor
	  *  \param dt Data type initial value.
	  */
	data_type(data_enum dt);

	/**  A data_type object is constructed with a type corresponding to
	  *  the argument string. If the argument string is not valid a
	  *  runtime_error exception is thrown.
	  *  \brief String constructor 
	  *  \param type Sring name of the data type.
	  */
	explicit data_type(const std::string& type);

	/**  Return the enumerated data type code.
	  *  \brief Data type accessor
	  *  \return Enumerated data type.
	  */
	int code(void) const;

	/**  Compare the data type with the argument.
	  *  \brief Compare types
	  *  \param c Data_type to which this is compared.
	  *  \return True if the data types is less than the argument data type.
	  */
	bool operator<(const data_type& c) const;

	/**  Compare the data type with the argument.
	  *  \brief Compare types
	  *  \param c Data_type to which this is compared.
	  *  \return True if the data types are equal.
	  */
	bool operator==(const data_type& c) const;

	/**  Compare the data type with the argument.
	  *  \brief Compare types
	  *  \param c Data_type to which this is compared.
	  *  \return True if the data types are unequal.
	  */
	bool operator!=(const data_type& c) const;

	/**  Set the data type as specified by the argument string. A 
	  *  runtime_error exception is thrown if the string is not a 
	  *  valid data type name.
	  *  \brief Set data type
	  *  \param type The data type name string.
	  */
	void set(const std::string& type);

	/**  Set the data type as specified by the argument string. A 
	  *  runtime_error exception is thrown if the string is not a 
	  *  valid data type name.
	  *  \brief Set data type
	  *  \param type The data type name string.
	  */
	void set(const char* type);

	/**  Set the data type as specified by a frame standard code.
	  *  \brief Set data type
	  *  \param type The frame standard type
	  */
	void set_frame_type(int ftype);

	/**  The data type word width in bytes is returned. 
	  *  \brief Get data word width.
	  *  \return Data type word width.
	  */
	int size(void) const;

	/**  Return an constant character pointer to the data type name
	  *  string.
	  *  \brief String accessor
	  *  \return Character pointer to the constant name string.
	  */
	const char* str(void) const;

	/**  Return the data type swap width (in bytes). For most data types
	  *  the swap size is equal to the data size. For complex numbers, the
	  *  swap size is the size of a single element (\e i.e. half the total
	  *  size).
	  *  \brief Get data word width.
	  *  \return Data element width.
	  */
	int swap_size(void) const;
	
    private:
	///  Enumerated data type.
	data_enum mDataType;
    };

    //==================================  Inline methods
    inline
    data_type::data_type(void)
      : mDataType(kByte1)
    {}

    inline
    data_type::data_type(data_enum type)
      : mDataType(type)
    {}

    inline
    data_type::data_type(const std::string& s) {
	set(s);
    }

    inline int
    data_type::code(void) const {
	return int(mDataType);
    }

    inline bool 
    data_type::operator<(const data_type& c) const {
	return int(mDataType) < int(c.mDataType);
    }

    inline bool 
    data_type::operator==(const data_type& c) const {
	return mDataType == c.mDataType;
    }

    inline bool 
    data_type::operator!=(const data_type& c) const {
	return mDataType != c.mDataType;
    }

} // namespace sends

#endif // !defined(SENDS_DATA_TYPE_HH)
