/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_DRIVER_API_HH
#define SENDS_DRIVER_API_HH

#include <string>
#include <vector>
#include "types.hh"

namespace sends {

    class loc_list;
    class user_data;

    /**  The sends driver api provides a standard interface for all data 
      *  sources.
      *  @memo Data input driver API.
      *  @version 1.0; last modified August 7, 2008
      *  @author  John Zweizig
      *  @ingroup sends_base
      */
    class driver_api {
    public:
	/**  The channel vector node contains all the information needed
	  *  to complete a read operation. For now, that is just the
	  *  channel index. In the future this may include stuff like
	  *  subchannel id, preallocated segments, etc.
	  */
	typedef chan_index channel_node;

	/**  Channel vector is used to pass a list of channels to be read
	  *  to the driver.
	  */
	typedef std::vector<channel_node> channel_vect;

	/**  Segment vector is used to hold an internal list of reserved
	  *  segments.
	  */
	typedef std::vector<seg_index> segment_vect;

    public:

	/**  Construct a sends device driver.
	  *  \brief Constructor.
	  */
	driver_api(void);

	/**  Destroy a sends device driver.
	  *  \brief Destructor.
	  */
	virtual ~driver_api(void);

	/**  Destroy a sends device driver.
	  *  \brief Destructor.
	  */
	virtual int debug(void) const;

	/**  Get the driver specific name.
	  *  \brief Driver name.
	  *  \return driver name
	  */
	virtual std::string name(void) const = 0;

	/**  Test whether this is an online device.
	  *  \brief test for online device.
	  *  \return true if devices is capable of online I/O.
	  */
	virtual bool online(void) const;

	/**  Get the gps time of currently available online data. The default
	  *  version returns 0 to indicate that no online data are available.
	  *  \brief Current online data time.
	  *  \return Start time of online data.
	  */
	virtual gps_type online_gps(void);

	/**  Test whether data are available from this device.
	  *  \brief Test for available data.
	  *  \return true if device has online data.
	  */
	virtual bool ready(void) const;

	/**  Read data from the specified path for the specified times. 
	  *  Channel names and segments they are to be read into are passed
	  *  as a list.
	  *  \brief Read channel ldata into segments. 
	  *  \param fl    Source descriptor list.
	  *  \param start Start gps time of data to be read.
	  *  \param end   End gps time of data to be read.
	  *  \param lst   List of channels and segment IDs.
	  *  \return Result code. 0 if all requested data were read,
	  *          otherwise -1.
	  */
	virtual int read_data(const loc_list& ll, gps_type start, 
			      gps_type end, const channel_vect& lst) = 0;

	/**  Decrement the use count for the specified channel in the 
	  *  driver's requested channel list. remove the channel if the
	  *  use count drops to zero.
	  *  \brief Configure to request the specified channel.
	  *  \param c   Index of channnel to be removed from the request list.
	  *  \return Result code. 0 if channel release succeded, otherwise -1.
	  */
	virtual int release_channel(chan_index c);

	/**  Add the specified channel to the driver's request list.
	  *  \brief Request the specified channel.
	  *  \param c   Index of channnel to be added to the request list.
	  *  \return Result code. 0 if channel request succeded, otherwise -1.
	  */
	virtual int request_channel(chan_index c);

	/**  Set the debug printout level.
	  *  \brief set the debug level.
	  *  \param lvl New debug level
	  */
	virtual void setDebug(int lvl);

	/**  Set the driver access level.
	  *  \brief Set the access level.
	  *  \param msk Access mask to be used by this driver.
	  */
	virtual void set_access(access_mask msk);

	/**  Return the level with which the user may access this driver.
	  *  \brief Test the access level.
	  *  \param usr Access mask available to this user.
	  */
	virtual access_mask test_access(const user_data& usr);

	/**  test whether data of the specified type are available in the 
	  *  specified path for the specified times.
	  *  \brief Read channel ldata into segments.
	  *  \param ll   Location list descriptor.
	  *  \param t    Start gps time of data to be read.
	  *  \return True if the specified data frame exists.
	  */
	virtual bool test_available(const loc_list& ll, 
				    gps_type t) const = 0;
    protected:
	/**  Segments are allocated to hold data for the specified channels
	  *  in the time range specified by (start, end). The segment numbers
	  *  are returned in the segment vector. Multiple segments may be 
	  *  reserved for a single channel if a single segment is too short 
	  *  to hold the data in the specified time epoch. get_segments may
	  *  not return a segment for a given channel if the data is already
	  *  available in a segment.
	  *  @brief Reserve segments for the specified channels.
	  *  @param start Start gps time.
	  *  @param end   End gps time.
	  *  @param cv    Vector of channel IDs to allocat segments for.
	  *  @param sv    Vector of segments.
	  *  @return Number of segments reserved.
	  */
	int get_segments(gps_type start, gps_type end, const channel_vect& cv,
			 segment_vect& sv);

	/**  Release a single segment. Unuse the segment and release the 
	  *  fill lock. This function reverses the reservation made by the
	  *  get_segments method.
	  *  @brief Release a single reserved segment.
	  *  @param sv Vector of segments.
	  */
	void release_segment(seg_index iseg) const;

	/**  Release segments in list.
	  *  @brief release a list of segments.
	  *  @param sv Vector of segments.
	  */
	void release_segments(segment_vect& sv);

    private:
	int         mDebug;
	access_mask mEnable;
    }; // class driver_api

    //==================================  Inline accessors
    inline int 
    driver_api::debug(void) const {
	return mDebug;
    }

    inline bool 
    driver_api::online(void) const {
	return false;
    }

    inline bool 
    driver_api::ready(void) const {
	return true;
    }

} // namespace sends

#endif // !defined(SENDS_DRIVER_API_HH)
