/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_DRIVER_TABLE_HH
#define SENDS_DRIVER_TABLE_HH

#include "driver_api.hh"
#include <string>
#include <vector>
#include <stdexcept>

namespace sends {

    /**  The sends driver table contains a list of loaded drivers.
      *  @memo Table of loaded drivers.
      *  @version 1.0; last modified August 7, 2008
      *  @author  John Zweizig
      *  @ingroup sends_base
      */
    class driver_table {
    public:

	/**  Static default driver index. 
	  */
	static const index_type default_driver = 0;

    public:

	/**  Construct a sends device driver.
	  *  \brief Constructor.
	  */
	driver_table(void);

	/**  Destroy a sends device driver.
	  *  \brief Destructor.
	  */
	~driver_table(void);

	/**  Add a driver to the table. Ownership of the driver class is
	  *  taken over by the driver table.
	  *  \brief Add a driver.
	  *  \param name Driver name.
	  *  \param d    driver api pointer.
	  */
	void add(const std::string& name, driver_api* d);

	/**  Add a driver to the table. The driver name is requested from
	  *  the driver class name() method. Ownership of the driver class 
	  *  is taken over by the driver table.
	  *  \brief Add a driver.
	  *  \param d    driver api pointer.
	  */
	void add(driver_api* d);

	/**  Get the driver type code.
	  *  \brief Get the type code.
	  *  \param name Nam of driver to be referenced.
	  *  \return driver type code.
	  */
	index_type get_type(const std::string& name) const;

	/**  Get a reference to the specified driver.
	  *  \brief Get driver reference.
	  *  \param d Index of requested driver.
	  *  \return Reference to the specified driver.
	  */
	driver_api& operator[](index_type d);

	/**  Get a constant reference to the specified driver.
	  *  \brief Get constant driver reference.
	  *  \param d Index of requested driver.
	  *  \return Constant reference to the specified driver.
	  */
	const driver_api& operator[](index_type d) const;

    private:
	typedef std::pair<std::string, driver_api*> table_node;
	typedef std::vector< table_node > driver_vect;

	driver_vect mVect;

    }; // class driver_table


    //==================================  External driver table
    extern driver_table TheDriverTable;

    //==================================  Inline methods.
    inline driver_api& 
    driver_table::operator[](index_type d) {
	if (d < mVect.size()) return *mVect[d].second;
	throw std::logic_error("Invalid driver index");
    }

    inline const driver_api& 
    driver_table::operator[](index_type d) const {
	if (d < mVect.size()) return *mVect[d].second;
	throw std::logic_error("Invalid driver index");
    }

} // namespace sends

#endif // !defined(SENDS_DRIVER_TABLE_HH)
