/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_EPOCH_HH
#define SENDS_EPOCH_HH

#include "types.hh"
#include <iosfwd>

namespace sends {

    /**  The %epoch class holds a time %epoch and performs various comparisons, 
      *  tests, etc operations.
      */
    class epoch {
    public:
	/** The end of all time = GPS 2G since the mayans got it wrong.
	  */
	static const gps_type gps_inf = 1999999999;
	static const epoch    ALL;
	static const epoch    null;

	/** Output style modes.
	 */
	enum epoch_style {
	    st_default, ///<  Default style \c (\<start>-\<end>)
	    st_delta,   ///<  Delta style   \c (\<start>:\<delta-t>)
	    st_trans    ///<  Default style with translated \c all and \c inf.
	};

    public:
	/**  Default %epoch constructor. The default constructor initializes
	  *  the %epoch to [0:gps_inf].
	  *  \brief Default constructor
	  */
	epoch(void);

	/**  Construct and initialize an %epoch.
	  *  \brief Epoch data constructor.
	  *  \param tStart Epoch start time.
	  *  \param tEnd   Epoch end time.
	  */
	epoch(gps_type tStart, gps_type tEnd=gps_inf);

	/**  Destroy an %epoch.
	  *  \brief Destructor.
	  */
	virtual ~epoch(void) {}

	/**  Test for adjacency. Two epochs are adjacent if they share a 
	  *  boundary time, \e i.e. if the start time of one %epoch equals 
	  *  the end time of the other.
	  *  \brief Test for adjacency
	  *  \param epic Epoch to be examined for adjacency.
	  *  \return True if epochs share a border time.
	  */
	bool adjacent(const epoch& epic) const;

	/**  Test for all times \c {0:inf}
	  *  \brief Test for ALL %epoch.
	  *  \return True if this instance covers all epochs.
	  */
	bool all(void) const;

	/**  Test whether the argument time lies inside this %epoch. Contained
	  *  times are \c start\<=\ t\ \<end.
	  *  \brief Test time in %epoch
	  *  \param t time to test.
	  *  \return true if time is contained.
	  */
	bool contains(gps_type t) const;

	/**  Return the %epoch duration in seconds.
	  *  \brief Duration.
	  *  \return Epoch duration in seconds.
	  */
	gps_type duration(void) const;

	/**  Display an epoch with the specified style.
	  *  \brief Formatted epoch contents.
	  *  \param out   Output stream.
	  *  \param style Printout style
	  *  \return refernce to modified stream.
	  */
	std::ostream& display(std::ostream& out, epoch_style style=st_default) const;

	/**  Epoch end GPS time.
	  *  \brief End time.
	  *  \return End GPS time seconds.
	  */
	gps_type end(void) const;

	/**  Merge two epochs by taking the earliest start time and the
	  *  latest end time. The result replaces the current instance.
	  *  The result of merging an empty epoch with a non-empty epoch is
	  *  identical to the non-empty epoch.
	  *  \brief Merge two epochs
	  *  \param epic Epoch to be merged with/into this instance.
	  */
	void merge(const epoch& epic);

	/**  Test for null %epoch \c {0:0}
	  *  \brief Test for null %epoch.
	  *  \return True if this instance covers null %epoch.
	  */
	bool none(void) const;

	/**  Compare starting time of two epochs.
	  *  \brief Compare %epoch starts.
	  *  \param epic Epoch instance to compare to this.
	  *  \return True if this start time is earlier than \a epic.
	  */
	bool operator<(const epoch& epic) const;

	/**  Compare two epochs for equality. Epochs are equal if they have
	  *  the same start and stop times or if they are both empty.
	  *  \brief Compare epochs.
	  *  \param epic Epoch instance to compare to this.
	  *  \return True if this instance is equal to \a epic.
	  */
	bool operator==(const epoch& epic) const;

	/**  Replace this instance with the "or" of this instance and 
	  *  the argument %epoch. If the instances are not adjacent,
	  *  the result is the earlier instance.
	  *  \brief Compute combined %epoch.
	  *  \param epic Epoch instance to "Or" with this.
	  *  \return Reference to the result instance.
	  */
	epoch& operator|=(const epoch& epic);

	/**  Replace this instance with the overlap of this instance and 
	  *  the argument %epoch. If there is no overlap between the argument
	  *  instances, the result is the empty %epoch {0, 0}.
	  *  \brief Compute %epoch overlap.
	  *  \param epic Epoch instance to and with this.
	  *  \return Reference to the result instance.
	  */
	epoch& operator&=(const epoch& epic);

	/**  Test for overlapping or adjacent epochs.
	  *  \brief Test for overlap
	  *  \param epic Instance to be compared to this.
	  *  \return true if the argument instance is adjacent to or 
	  *          overlapping with this instance.
	  */
	bool overlap(const epoch& epic) const;

	/**  Test for overlapping epochs.
	  *  \brief Test for overlap
	  *  \param epic Instance to be compared to this.
	  *  \return true if the argument overlaps this instance.
	  */
	bool overlapx(const epoch& epic) const;

	/**  Parse an %epoch string and return the %epoch value. A progress 
	  *  pointer is also returned.
	  *  \brief Parse %epoch from a character string.
	  *  \param str  Pointer to the string to be parsed.
	  *  \param pptr progress pointer address.
	  */
	static epoch parse(const char* str, const char** pptr=0);

	/**  Epoch start GPS time.
	  *  \brief Start time.
	  *  \return Start GPS time seconds.
	  */
	gps_type start(void) const;
    private:
	gps_type _start;
	gps_type _end;
    };

    //==================================  output function
    std::ostream& operator<<(std::ostream& out, const epoch& epic);

    //==================================  inline methods.
    inline
    epoch::epoch(void)
	: _start(0), _end(gps_inf) 
    {}

    inline
    epoch::epoch(gps_type tStart, gps_type tEnd)
	: _start(tStart), _end(tEnd) 
    {}

    inline bool
    epoch::adjacent(const epoch& epic) const {
	return _start == epic._end || _end == epic._start;
    }

    inline bool
    epoch::all(void) const {
	return !_start && _end == gps_inf;
    }

    inline bool
    epoch::contains(gps_type t) const {
	return t < _end && t >= _start;
    }

    inline gps_type 
    epoch::duration(void) const {
	return _end - _start;
    }

    inline gps_type  
    epoch::end(void) const {
	return _end;
    }

    inline void
    epoch::merge(const epoch& epic) {
	if (epic.none()) return;
	if (none()) {
	    _start = epic._start;
	    _end   = epic._end;
	} else {
	    if (epic._start < _start) _start = epic._start;
	    if (epic._end   > _end)   _end   = epic._end;
	}
    }

    inline bool
    epoch::none(void) const {
	return _start >= _end;
    }

    inline bool  
    epoch::operator<(const epoch& epic) const {
	return _start < epic._start;
    }

    inline bool  
    epoch::operator==(const epoch& epic) const {
	return (_start == epic._start && _end == epic._end) ||
	       (none() && epic.none());
    }

    inline epoch&
    epoch::operator|=(const epoch& epic) {
	if (overlap(epic)) {
	    if (epic._start < _start) _start = epic._start;
	    if (epic._end   > _end  )   _end = epic._end;
	}
	return *this;
    }

    inline epoch&
    epoch::operator&=(const epoch& epic) {
	if (overlapx(epic)) {
	    if (epic._start > _start) _start = epic._start;
	    if (epic._end   < _end  ) _end = epic._end;
	}
	else {
	    _start = 0;
	    _end = 0;
	}
	return *this;
    }

    inline bool
    epoch::overlap(const epoch& epic) const {
	return _start <= epic._end && epic._start <= _end;
    }

    inline bool
    epoch::overlapx(const epoch& epic) const {
	return _start < epic._end && epic._start < _end;
    }

    inline gps_type  
    epoch::start(void) const {
	return _start;
    }

}  //   namespace sends.

#endif // !defined(SENDS_EPOCH_HH)
