/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "epoch.hh"
#include <map>
#include <string>
#include <stdexcept>

namespace sends {

    /**  The epoch table keeps a map of defined epochs.
      *  \brief Epoch lookup table class.
      *  \author John.zweizig
      *  \version $Id$
      */
    class epoch_table {
    public:
	/**  Construct an empty table.
	  *  \brief Default constructor.
	  */
	epoch_table(void);

	/**  Destroy an %epoch table.
	  *  \brief Destructor.
	  */
	~epoch_table(void);

	/**  Add an %epoch to the table.
	  *  \brief Add an %epoch.
	  *  \param name Epoch name.
	  *  \param epic Epoch time range.
	  */
	void add(const std::string& name, const epoch& epic);

	/**  Test whether the specified %epoch name is define in the table.
	  *  \brief Test whether an %epoch is defined.
	  *  \param name Epoch name to test.
	  *  \return True if the named %epoch is defined.
	  */
	bool defined(const std::string& name) const;

	/**  Print out all the epochs defined in this table to the specified
	  *  output stream.
	  *  \brief Print defined epochs.
	  *  \param out Output stream.
	  *  \return Reference to output stream.
	  */
	std::ostream& display(std::ostream& out) const;

	/**  List all the defined epochs defined in this table to the specified
	  *  output stream in the format "<name>=<epoch>".
	  *  \brief List defined epochs.
	  *  \param out Output stream.
	  *  \return Reference to output stream.
	  */
	std::ostream& list(std::ostream& out) const;

	/**  Fetch an epoch from the table.
	  *  \brief Fetch %epoch.
	  *  \param name Name of %epoch to be fetched.
	  *  \return Reference to named %epoch
	  */
	const epoch& operator[](const std::string& name) const;
    private:
	///  Epoch map data type
	typedef std::map<std::string, epoch> epoch_map;
	///  Epoch map iterator.
	typedef epoch_map::const_iterator    const_epoch_iter;
    private:
	epoch_map  mTable;
    };

    //==================================  inline methods.
    inline bool
    epoch_table::defined(const std::string& name) const {
	return mTable.find(name) != mTable.end();
    }

    inline const epoch&
    epoch_table::operator[](const std::string& name) const {
	const_epoch_iter iter =  mTable.find(name);
	if (iter != mTable.end()) return iter->second;
	throw std::runtime_error("Epoch name not found"); 
    }

    extern epoch_table Epoch_Table;

} //namespace sends
