/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_FETCH_STAGE_HH
#define SENDS_FETCH_STAGE_HH

#include "types.hh"
#include "seda_stage.hh"

namespace sends {

    class frame_loc;

    /**  The fetch_stage class provides a minimal template for a SEDA 
      *  stage thread pool. It is based on the seda_stage class and provides
      *  a stage with no filter processing. The constructor, destructor and 
      *  filter methods are implemented. The filter method immediately passes 
      *  the transactions to its output queue (the input queue of the next 
      *  stage in the pipeline.
      *  @brief Data fetch SEDA stage.
      *  @author John Zweizig
      *  @version 1.0; Last modified December 28, 2007
      *  @ingroup sends_stage
      */
    class fetch_stage : public seda_stage {
    public:
	/**  Construct a fetch_stage object. The arguments indicate the 
	  *  number of threads in the pool and the number of input transactions
	  *  that may be queue simultaneously (note that the pipline gets 
	  *  indigestion only after more than (nTrans + nThread) trasactions
	  *  need to execute in this stage.
	  *  @brief Stage template constructor.
	  *  @param nThread Number of threads to run simultaneously in the pool.
	  *  @param nTrans  Number of transactions that may be queued.
	  */
	fetch_stage(int nThread, int nTrans);

	/**  Stage template destructor is null by default. It is mostly
	  *  useful to remove resources allocated by the stages.
	  */
	virtual ~fetch_stage(void);

	/**  The fetch_stage filter method detemines what actions must be 
	  *  taken for the transaction type. It then calls find_data or
	  *  passes the transaction to the next stage as appropriate.
	  *  @brief Transaction filter for this stage.
	  *  @param trans Pointer to the current transaction 
	  */
	void filter(transaction* trans);

	/**  The find_data method builds a list of segments with the data 
	  *  requested for the transaction. If data are not currently 
	  *  available in segments buffers they are read in using the I/O
	  *  driver specfied by the frame location block. This is 
	  *  accomplished by iterating over the following steps:
	  *  \li Build up a source histogram using the source_histo class.
	  *  \li Select the most used source - release other entries from 
	  *      the histogram. If no sources are found for any of the 
	  *      current unsourced channels, mark the remaining channels
	  *      as not available.
	  *  \li Build a list of channels available from the selected source.
	  *  \li Read in all data from the specified source 
	  *
	  *  @brief Search for a data source and read in data.
	  *  @param trans Pointer to the current transaction 
	  */
	void find_data(transaction* trans);

	/**  Return a constant pointer to a null-terminated character vector 
	  *  containing the name of the stage.
	  *  @brief Stage name.
	  *  @return Constant pointer to the stage name string.
	  */
	const char* stage_name(void) const;

    private:
    };
}

#endif // !defined(SENDS_FETCH_STAGE_HH)
