/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_FONL_DRIVER_HH
#define SENDS_FONL_DRIVER_HH

#include "frame_driver.hh"
#include "loc_list.hh"

namespace sends {

    class src_update;


    /**  NDS2 frame input driver for online file systems.
      *  \brief Online frame input driver.
      *  \ingroup sends_drive
      *  \version 1.1; last modified November 27, 2008.
      *  \author John Zweizig
      */
    class fonl_driver : public frame_driver {
    public:

	/**  Construct a posix file-system driver.
	  *  \brief Constructor
	  */
	fonl_driver(const std::string& dname);

	/**  Destroy a posix file-system driver.
	  *  \brief Destructor
	  */
	~fonl_driver(void);

	/**  Find the updater and set the pointer.
	  *  \brief Find the drivers updater.
	  *  \return True if valid updater.
	  */
	bool find_update(void);

	/**  Get the driver specific name ("file").
	  *  \brief get the driver name.
	  *  \return driver name.
	  */
	std::string name(void) const;

	/**  Get online status flag.
          *  return true for online drivers.
          *  \brief online flag
          */
	bool online(void) const;

	/**  Get the gps time of currently available online data. If no data 
	  *  segment is available a new segment is read and the time is 
	  *  returned.
	  *  \brief Current online data time.
	  *  \return Start time of online data.
	  */
	gps_type online_gps(void);

	/**  Read data from the specified path for the specified times. 
	  *  Channel names and segments they are to be read into are passed
	  *  as a list.
	  *  \brief Read channel ldata into segments. 
	  *  \param type  Source frame list.
	  *  \param path  Source descriptor path field.
	  *  \param start Start gps time of data to be read.
	  *  \param end   End gps of data to be read.
	  *  \param lst   List of channels and segment IDs.
	  *  \return Result code. 0 if all requested data were read,
	  *          otherwise -1.
	  */
	virtual int read_data(const loc_list& type, gps_type start, 
			      gps_type dt, const channel_vect& lst);

	/**  Set the driver name. The driver type ("fonl:") is prepended to
	  *  the name given.
	  */
	void set_name(const std::string& dname);

	/**  Set the source updater name.
	 */
	void set_update(const std::string& uname);

    private:
	std::string mName;
	std::string mUpdName;
	src_update* mUpdate;
    };

    //==================================  Inline methods
    inline std::string 
    fonl_driver::name(void) const {
	return mName;
    }

    inline bool
    fonl_driver::online(void) const {
        return true;
    }

}  // namspace sends

#endif   // !defined(SENDS_FONL_DRIVER_HH)
