/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_FRAME_DRIVER_HH
#define SENDS_FRAME_DRIVER_HH

#include "driver_api.hh"
#include "framecpp/IFrameStream.hh"

namespace sends {
    class loc_list;

    /**  The sends frame driver is an abstract sends driver with some frame
      *  I/O functionality included as protected methods.
      *  @memo Frame input driver.
      *  @ingroup sends_drive
      *  @version 1.0; last modified November 24, 2008
      *  @author  John Zweizig
      */
    class frame_driver : public driver_api {
    public:
	/** FrameCPP input stream.
	  */
	typedef FrameCPP::IFrameStream frame_stream;

    public:

	/**  Construct a sends device driver.
	  *  \brief Constructor.
	  */
	frame_driver(void);

	/**  Destroy a sends device driver.
	  *  \brief Destructor.
	  */
	virtual ~frame_driver(void);

	/**  Get the driver specific name.
	  *  \brief Driver name.
	  *  \return driver name
	  */
	virtual std::string name(void) const = 0;

	/**  Read data from the specified path for the specified times. 
	  *  Channel names and segments they are to be read into are passed
	  *  as a list.
	  *  \brief Read channel ldata into segments. 
	  *  \param ll    Source descriptor list.
	  *  \param start Start gps time of data to be read.
	  *  \param end   End gps time of data to be read.
	  *  \param lst   List of channels and segment IDs.
	  *  \return Result code. 0 if all requested data were read,
	  *          otherwise -1.
	  */
	virtual int read_data(const loc_list& ll, gps_type start, 
			      gps_type end, const channel_vect& lst) = 0;

	/**  test whether data of the specified type are available in the 
	  *  specified path for the specified times.
	  *  \brief Read channel ldata into segments.
	  *  \param fl   Frame location descriptor list.
	  *  \param t    Start gps time of data to be read.
	  *  \return True if the specified data frame exists.
	  */
	virtual bool test_available(const loc_list& ll, 
				    gps_type t) const;
    protected:
	/**  Read data from the specified frame stream into the allocated
	  *  segments.
	  *  \brief Getspecified data from frame.
	  *  \param str    FrameCPP input stream.
	  *  \param segv   Segment ID vector.
	  *  \param iFrame Frame number withing file.
	  *  \param gps    Frame start gps (in case it wasn't specified).
	  *  \return Zero if all requested data were read successfully.
	  */
	virtual int read_frame(frame_stream& str, segment_vect& segv, 
			       int iFrame=0, gps_type gps=0) const;
    }; // class frame_driver

} // namespace sends

#endif // !defined(SENDS_FRAME_DRIVER_HH)
