/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_FRAME_LOC_HH
#define SENDS_FRAME_LOC_HH

#include "types.hh"
#include <string>
#include <iosfwd>

namespace sends {

    /**  The frame_loc class contains information about a set of consecutive 
      *  frames of a given type in a specified area. The frame source type
      *  consists of the first two fields of the frame file name, i.e. the
      *  site code and the frame type, e.g. \c "H-RDS_R_L1". The source path
      *  may contain escaped format descriptions that are replaced by the
      *  start GPS of a frame when the path is resolved. For example, 
      *  frame_path() would replace the escaped format sequence in 
      *  \c "/archive/frames/S5/L0/LLO/L-R-%3d" with the first 3 digits of the
      *  gps time of the frame being resolved.
      *  @brief Location of a series of frame files.
      *  @author John Zweizig
      *  @version 1.0; Last Modified January 22, 2008
      *  @ingroup sends_base
      */
    class frame_loc {
    public:

	/**  Frame data input driver type.
	 */
	typedef index_type driver_type;
 
    public:
	/**  Construct an empty frame location descriptor.
	  *  @brief Empty frame location constructor.
	  */
	frame_loc(void);

	/**  Construct a frame location descriptor.
	  *  @brief Constructor.
	  *  @param location Path to frame archive.
	  *  @param prefix   Frame type string.
	  *  @param start    Start GPS of first frame in series.
	  *  @param stop     End GPS of last frame in series.
	  *  @param dT       Time duration of all but the last frame.
	  *  @param driv     Driver type.
	  */
	frame_loc(const std::string& location, const std::string& prefix, 
		  gps_type start, gps_type stop, gps_type dT, 
		  driver_type driv=0);

	/**  Destroy a frame location descriptor.
	  *  @brief Frame location destructor.
	  */
	~frame_loc(void);

	/**  Get the latest gps time that is equal to 
	  *  \c start+N*duration
	  *  and not after the specified time,
	  *  @brief Get an aligned time.
	  *  @param t0 Reference time.
	  *  @return Aligned gps time not after t0.
	  */
	gps_type align_gps(gps_type t0) const;

	/**  Get the driver type code for this frame location.
	  *  @brief  Get the driver type.
	  *  @return Driver type code.
	  */
	driver_type driver(void) const;

	/**  Get the duration parameter.
	  *  @brief  Get the duration parameter.
	  *  @return Duration time.
	  */
	gps_type duration(void) const;

	/**  Dump the location entry
	  *  @brief  Dump the location.
	  *  @param out Output stream.
	  *  @return Output stream reference.
	  */
	std::ostream& dump(std::ostream& out) const;

	/**  Return the reconstructed path string with gps substitutions.
	  *  The full file path is: \c "<path>/<prefix>-<gps>-<duration>.gwf".
	  *  The following substitution characters may be used in the file 
	  *  path string. 
	  *  - \%\<n\>d Number of high-order GPS digits.
	  *  - \%\<n\>D Day in GPS year.
	  *  - \%\<n\>h GPS hour.
	  *  - \%\<n\>r Remove n low-order decimal digits.
	  *  - \%\<n\>Y GPS year (\e e.g. 2011).
	  *
	  *  @brief  Return path string with gps substitutions. 
	  *  @param  t GPS time to be substituted into path.
	  *  @param  limit Limit path name to the %frame_loc duration.
	  *  @return Frame prefix string.
	  */
	std::string file_path(gps_type t, bool limit=true) const;

	/**  Get the end time of the frame containing the specified time.
	  *  @brief  Get the frame end time.
	  *  @param t Time contained in frame.
	  *  @return End time of frame.
	  */
	gps_type frame_end(gps_type t) const;

	/**  Get the start time of the frame containing the specified time.
	  *  @brief  Get the frame start time.
	  *  @param t Time contained in frame.
	  *  @return Start time of frame.
	  */
	gps_type frame_start(gps_type t) const;

	/**  Test whether the specified gps is included in the valid time range
	  *  for this location.
	  *  @brief Test if time is in range.
	  *  @param gps Time to be checked in GPS seconds.
	  *  @return True if gps is in range covered by frame_loc.
	  */
	bool in_range(gps_type gps) const;

	/**  Return memory used by this frame_loc;
	  *  \brief Memory use.
	  *  \return Number of bytes of memory used.
	  */
	index_type mem_use(void) const;

	/**  Merge an argument frame locator into this instance by extending
	  *  either the start time or end time. The merge will not take place
	  *  if the time ranges were not contiguous or if any other parameters
	  *  \em i.e. the location, prefix, duration, driver or archive type 
	  *  differ. The merge will also be considered successful if the 
	  *  argument %frame_loc is completely contained inside this instance 
	  *  with the correct alignment.
	  *  \brief Merge a %frame_loc.
	  *  \param fl Reference to the %frame_loc instance to be merged 
	  *            into this.
	  *  \return True if the merge was successful.
	  */
	bool merge(const frame_loc& fl);

	/**  On_tape tests whether the file is on spinning media or on tape.
	  *  The frame containing the specified gps time (t0) is tested for
	  *  holding a non-zero space. 
	  */
	bool on_tape(gps_type t0) const;

	/**  On_tape tests whether the file is on spinning media or on tape.
	  *  There are two versions
	  */
	bool on_tape(gps_type t0, gps_type t1) const;

	/**  Compare ordering of two %frame_loc segments. The comparison
	  *  is dependent only on the start times.
	  *  \brief Sequence order comparison.
	  *  \param x reference to %frame_loc to be compared to this.
	  *  \return True if the start time of this instance is before the
	  *           start of the argument %frame_loc.
	  */
	bool operator<(const frame_loc& x) const;

	/**  Return the driver-dependent path string.
	  *  @brief Return path string.
	  *  @return Path description string.
	  */
	const std::string& path(void) const;

	/**  Return the frame site and type string covered by this frame_loc.
	  *  @brief Source prefix.
	  *  @return Frame prefix string.
	  */
	const std::string& prefix(void) const;

	/**  Set the driver for frames in this location.
	  *  \brief Set the driver 
	  *  \param drv Driver id.
	  */
	void set_driver(driver_type drv);

	/**  Set the file system path string for this frame set.
	  *  \brief Set the path 
	  *  \param loc New path string
	  */
	void set_path(const std::string& loc);

	/**  Set the start time of the first frame in this location.
	  *  \brief Set the start time of data at this location 
	  *  \param t Data start time
	  */
	void set_start(gps_type t);

	/**  Set the end time of the last frame in this location.
	  *  \brief Set the end time of data at this location 
	  *  \param t Data end time
	  */
	void set_stop(gps_type t);

	/**  Return the start time of the first frame in this location.
	  *  \brief Start time of data at this location 
	  *  \return Data start time
	  */
	gps_type start(void) const;

	/**  Return the end time of the last frame in this location.
	  *  \brief Start time of data at this location 
	  *  \return Data end time
	  */
	gps_type stop(void) const;

	/**  Swap contents of the frame locator with the argument locator.
	  *  Awesome for sorts and things like that.
	  *  \brief Swap two locators.
	  *  \param x Locator with contents to swapped with this.
	  */
	void swap(frame_loc& x);

    private:
	std::string mLocation;
	std::string mPrefix;
	gps_type    mStart;
	gps_type    mStop;
	gps_type    mDuration;
	driver_type mDriver;
    };

    //==================================  Inline methods
    inline gps_type
    frame_loc::align_gps(gps_type t0) const {
	if (t0 < mStart) return 0;
	return t0 - (t0 - mStart)%mDuration;
    }

    inline frame_loc::driver_type 
    frame_loc::driver(void) const {
	return mDriver;
    }

    inline gps_type 
    frame_loc::duration(void) const {
	return mDuration;
    }

    inline gps_type 
    frame_loc::frame_end(gps_type t) const {
	gps_type end = align_gps(t) + mDuration;
	if (end > mStop) end = mStop;
	return end;
    }

    inline gps_type 
    frame_loc::frame_start(gps_type t) const {
	return align_gps(t);
    }

    inline bool
    frame_loc::in_range(gps_type gps) const {
	return gps >= mStart && gps < mStop;
    }

    inline index_type
    frame_loc::mem_use(void) const {
	return sizeof(frame_loc) + mLocation.size() + mPrefix.size();
    }

    inline bool
    frame_loc::operator<(const frame_loc& x) const {
	return (mStart < x.mStart);
    }

    inline const std::string& 
    frame_loc::path(void) const {
	return mLocation;
    }

    inline const std::string& 
    frame_loc::prefix(void) const {
	return mPrefix;
    }

    inline gps_type 
    frame_loc::start(void) const {
	return mStart;
    }

    inline gps_type 
    frame_loc::stop(void) const {
	return mStop;
    }

} // namespace sends

#endif  // !defined(SENDS_FRAME_LOC_HH)
