/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_FRAME_SOURCE_HH
#define SENDS_FRAME_SOURCE_HH

#include <map>
#include <vector>
#include <iosfwd>
#include "types.hh"
#include "loc_list.hh"


namespace sends {

    /**  The frame_source class contains several lists of frame locations,
      *  one list for each archive level. The lists contain all information
      *  necessary to decide on a most easily available source and to get the
      *  driver and file name for the source. The lists are populated from 
      *  one or more ascii files read by the read() method. The read() method
      *  documentation describes these files.
      *  @brief Frame data location lists.
      *  @author John Zweizig
      *  @version $Id: frame_source.hh 1622 2016-02-21 01:48:49Z john.zweizig@LIGO.ORG $
      *  @ingroup sends_base
      */
    class frame_source {
    public:

	/**  Archive level enumeration type .
	  */
	typedef loc_list::arch_type arch_type;

	/** Define frame source map type.
	  */
	typedef std::map<std::string, loc_list> frame_map;

	/** Define frame source map node type.
	  */
	typedef frame_map::value_type frmap_node;

	/** Define constant frame source map iterator type.
	  */
	typedef frame_map::const_iterator const_frmap_iter;

	/** Define frame map iterator type.
	  */
	typedef frame_map::iterator frmap_iter;

	/** Define a GPS segment type
	  */
	typedef std::pair<gps_type, gps_type> gpsseg_type;

	/**  Constant used to indicate infinite time.
	  */
	static const gps_type inf_gps=1999999999;

    public:
	/**  Construct a Frame Source database. All the lists are constructed
	  *  but left empty.
	  *  @brief Default frame_source database constructor.
	  */
	frame_source(void);

	/**  Frame source database destructor.
	  *  @brief Destructor.
	  */
	~frame_source(void);

	/**  Dump the contents of all the frame source lists.
	  *  @brief Dump the lists.
	  *  @param out Output stream.
	  *  @param level Verbosity level: 0=no data, 1=start and stop 
	  *               times only, 2=full structures.
	  *  @return Reference to the output stream.
	  */
	std::ostream& dump(std::ostream& out, int level=1) const;

#ifdef LOC_LIST_LOCKED_FINDS
	/**  Find a frame location for the specified type and gps time. The 
	  *  search iterates over all the lists in the order online, cache,
	  *  archive, offline. The first valid location found in this sequence 
	  *  is returned. If the requested frame locator is found, the location
	  *  list for the  requested type is locked.
	  *  @brief Find a frame location.
	  *  @param type Frame type prefix string.
	  *  @param gps GPS time to be found.
	  *  @return Constant reference to the frame location.
	  *  @exception runtime_error is thrown if no frame is found.
	  */
	const frame_loc* find(const std::string& type, gps_type gps) const;
#endif
	
	/**  Find a frame location for the specified type and gps time. The 
	  *  search iterates over all the lists in the order online, cache,
	  *  archive, offline. The first valid location found in this sequence 
	  *  is returned. If the requested frame locator is found, the location
	  *  list for the  requested type is locked.
	  *  @brief Find a frame location.
	  *  @param type Frame type prefix string.
	  *  @param gps GPS time to be found.
	  *  @param fl Frame loc instance to which the result is written
	  *  @return True if valid frame locater was found.
	  */
	bool find(const std::string& type, gps_type gps, frame_loc& fl) const;

#ifdef LOC_LIST_LOCKED_FINDS
	/**  Find a frame location for the specified archive level, type and 
	  *  gps time. If the requested frame locator is found, the location
	  *  list for the  requested type is locked.
	  *  @brief Find a frame location.
	  *  @param a    Frame archive level code.
	  *  @param type Frame type prefix string.
	  *  @param gps  GPS time to be found.
	  *  @return Constant pointer to the frame location or NULL.
	  */
	const frame_loc* find(arch_type a, const std::string& type,
			      gps_type gps) const;

	/**  Find a frame location for the specified archive level, type and 
	  *  gps time.If the requested frame locator is found, the location
	  *  list for the  requested type is locked.
	  *  @brief Find a frame location.
	  *  @param arch Frame archive level code.
	  *  @param type Frame type prefix string.
	  *  @param gps GPS time to be found.
	  *  @return Constant pointer to the frame location or NULL.
	  */
	frame_loc* find(arch_type arch, const std::string& type, gps_type gps);
#endif
	
	/**  Find a frame location for the specified archive level, type and 
	  *  gps time. If the requested frame locator is found, the location
	  *  list for the  requested type is locked.
	  *  @brief Find a frame location.
	  *  @param a    Frame archive level code.
	  *  @param type Frame type prefix string.
	  *  @param gps  GPS time to be found.
	  *  @return True if a valid frame_loc instance was found
	  */
	bool find(arch_type a, const std::string& type, gps_type gps,
		  frame_loc& fl) const;

	/**  Find a frame location for the specified archive level, type and 
	  *  gps time.If the requested frame locator is found, the location
	  *  list for the  requested type is locked.
	  *  @brief Find a frame location.
	  *  @param arch Frame archive level code.
	  *  @param type Frame type prefix string.
	  *  @param gps  GPS time to be found.
	  *  @param fl   Frame locater instance to receive found data.
	  *  @return True of a valid frame locater was found.
	  */
	bool find(arch_type arch, const std::string& type, gps_type gps,
		  frame_loc& fl);

#ifdef LOC_LIST_LOCKED_FINDS
	/**  Find the last frame location object for the specified archive 
	  *  level and prefix string. If the requested frame locator is found, 
	  *  the location list for the requested loc_list is locked.
	  *  @brief Find a frame location.
	  *  @param a   Frame archive level code.
	  *  @param pfx Frame type prefix string.
	  *  @return Constant pointer to the frame location or NULL.
	  */
	frame_loc* find_back(arch_type a, const std::string& pfx);

	/**  Find the last frame location object for the specified archive 
	  *  level and prefix string. If the requested frame locator is found, 
	  *  the location list for the requested loc_list is locked.
	  *  @brief Find a frame location.
	  *  @param a   Frame archive level code.
	  *  @param pfx Frame type prefix string.
	  *  @return Constant pointer to the frame location or NULL.
	  */
	const frame_loc* find_back(arch_type a, const std::string& pfx) const;
#endif

	/**  Find the last frame location object for the specified archive 
	  *  level and prefix string. If the requested frame locator is found, 
	  *  the location list for the requested loc_list is locked.
	  *  @brief Find a frame location.
	  *  @param a   Frame archive level code.
	  *  @param pfx Frame type prefix string.
	  *  @param f   Frame location instance found.
	  *  @return True if frameloc is valid.
	  */
	bool find_back(arch_type a, const std::string& pfx, frame_loc& f) const;

	/**  Find the earliest location entry for the specified prefix 
	  *  string. If the requested frame locator is found, 
	  *  the location list for the requested loc_list is read locked.
	  *  @brief Find a frame location.
	  *  @param pfx Frame type prefix string.
	  *  @return Constant pointer to the frame location or NULL.
	  */
	const loc_list* find_first(const std::string& pfx) const;

#ifdef LOC_LIST_LOCKED_FINDS
	/**  Find the first location entry for the specified archive
	  *  level and prefix string. If the requested frame locator is found, 
	  *  the location list for the requested loc_list is read locked.
	  *  @brief Find a frame location.
	  *  @param a   Frame archive level code.
	  *  @param pfx Frame type prefix string.
	  *  @return Constant pointer to the frame location or NULL.
	  */
	const frame_loc* find_first(arch_type a, const std::string& pfx) const;
#endif

	/**  Find the first location entry for the specified archive
	  *  level and prefix string. If the requested frame locator is found, 
	  *  the location list for the requested loc_list is read locked.
	  *  @brief Find a frame location.
	  *  @param a   Frame archive level code.
	  *  @param pfx Frame type prefix string.
	  *  @param fl  Frame location instance to receive valid location. 
	  *  @return True if a valid frame locater was found.
	  */
	bool find_first(arch_type a, const std::string& pfx,
			frame_loc& fl) const;

	/**  Find the limits of the gap in frames of the given type around the
	  *  specified GPS time. The gap interval is returned to the argument
	  *  gps-segment. If no gap is found the gap argument contains a zero
	  *  length field.
	  *  \brief Find a gap in the frames.
	  *  \param pfx Frame type prefix string.
	  *  \param gps Point at which a gap is to be measured.
	  *  \param gap Gap limit values.
	  */
	void find_gap(const std::string& pfx, gps_type gps, 
		      gpsseg_type& gap) const;

	/**  Find the limits of the gap in frames of the given type around the
	  *  specified GPS time. The gap interval is returned to the argument
	  *  gps-segment. If no gap is found the gap argument contains a zero
	  *  length field.
	  *  \brief Find a gap in the frames.
	  *  \param pfx Frame type prefix string.
	  */
	void find_gap(const std::string& pfx, arch_type a, gps_type gps, 
		      gpsseg_type& gap) const;

	/**  Find the frame location object for the specified prefix with 
	  *  the latest end time. If the requested frame locator is found, 
	  *  the location list for the requested loc_list is read locked.
	  *  @brief Find a frame location.
	  *  @param pfx Frame type prefix string.
	  *  @return Constant pointer to the frame location or NULL.
	  */
	const loc_list* find_last(const std::string& pfx) const;

	/**  Find a frame location for the specified type and gps time. The 
	  *  search iterates over all the lists in the order online, cache,
	  *  archive, offline. If valid data are found the method returns true.
	  *  @brief Test for data at the specified time.
	  *  @param type Frame type prefix string.
	  *  @param gps GPS time to be found.
	  *  @return true if data are found.
	  */
	bool gps_valid(const std::string& type, gps_type gps) const;

	/**  Insert the specified frame location into the appropriate list. 
	  *  The list is selected by the archive type field in the frame_loc
	  *  and the frame type (key) is provided by the prefix field of the
	  *  frame_loc.
	  *  @brief Insert frame location int a source list.
	  *  @param arch List archive type.
	  *  @param loc  Frame_loc block to be inserted into the list.
	  */
	void insert(arch_type arch, const frame_loc& loc);

	/**  Return a constant pointer to the loc_list matching the 
	  *  specified frame type gps time. A gps time of 0 matches only
	  *  an online list.
	  *  A NULL pointer is returned if an appropriate list is not 
	  *  found.
	  *  \brief Reference a specified location list.
	  *  \param pfx  Frame type (prefix) of list to be referenced.
	  *  \param gps  GPS time of requested data.
	  *  \return Constant pointer to the specified list or NULL.
	  */
	const loc_list* list_ptr(const std::string& pfx, gps_type gps) const;

	/**  Return a constant pointer to the lock list for the specified 
	  *  archive level and frame type. If the specified list is not
	  *  found, a NULL pointer is returned.
	  *  \brief Reference a specified location list.
	  *  \param arch Archive level of list to be referenced.
	  *  \param pfx  Frame type (prefix) of list to be referenced.
	  *  \return Constant pointer to the specified list or NULL.
	  */
	const loc_list* list_ptr(arch_type arch, const std::string& pfx) const;

	/**  Return a constant to the loc_list for the specified archive
	  *  level and frame type. If the specified list is not found,
	  *  a NULL pointer is returned.
	  *  \brief Reference a specified location list.
	  *  \param arch Archive level of list to be referenced.
	  *  \param pfx  Frame type (prefix) of list to be referenced.
	  *  \return Constant pointer to the specified list or NULL.
	  */
	loc_list* list_ptr(arch_type arch, const std::string& pfx);

	/**  Get total memory use.
	  *  \brief memory usage.
	  *  \return total bytes used.
	  */
	index_type mem_use(void) const;

	/// range vector type
	typedef std::vector<std::pair<std::string, epoch> > range_vect;

	/**  Get a list of valid ranges for each data type.
	  */
	void range_list(range_vect& range) const;
	
	/**  Read a list of frame locations from a flat ascii file and use 
	  *  them to fill in the frame source tables. The input file 
	  *  contains a 7-column table with one row per line. Comments may
	  *  be added by including a '#' at the start of the line or after 
	  *  all text on the line, and lines may be continued by terminating
	  *  the line with a backslash. The columns are separated by one or
	  *  more white-space (space or tab) characters and contain the 
	  *  following information.
	  *  <table>
	  *  <tr><td>Column</td><td>Contents</td></tr>
	  *  <tr><td>1</td><td>Frame site and type, e.g "H-RDS_R_L1"</td></tr>
	  *  <tr><td>2</td><td>Start GPS of first frame</td></tr>
	  *  <tr><td>3</td><td>End GPS of last frame</td></tr>
	  *  <tr><td>4</td><td>Time duration of all frame in list</td></tr>
	  *  <tr><td>5</td><td>Archive type, i.e. "online", "cache", "archive",
	  *                    or offline</td></tr>
	  *  <tr><td>6</td><td>Driver type, i.e. "local", "qfs", "sends",
	  *                    "smem" or "unknown". "-" may be specified for a 
	  *                    arch-dependent default.</td></tr>
	  *  <tr><td>7</td><td>Source path. The path for local and qfs data
	  *                    may contain fields derived from the GPS time,
	  *                    e.g. "%d" will be replaced by the gps time and
	  *                    "%3d" ill be replace by the first 3 decimal 
	  *                    digits of the gps</td></tr>
	  *  </table>
	  *  If the fourth column (duration) is zero the specified path is 
	  *  scanned for valid frame files. The frames found by the scan that 
	  *  occur between the specified start and end GPS times are added to 
	  *  the source list.
	  *  \brief Read a frame source file.
	  *  \param file Path to source file.
	  */
	void read(const std::string& file);

	/**  Scan directories for frame files that match the location template
	  *  used by a previous scan. The frame source must have been scanned 
	  *  previously. 
	  *  \brief Rescan directories.
	  *  \exception runtime_error if the frame source was not previously 
	  *   scanned (\e i.e. if the scan path is not set).
	  *  \param arch   Location template string.
	  *  \param prefix Frame file prefix.
	  */
	void rescan(loc_list::arch_type arch, const std::string& prefix);

	/**  Scan directories for frame files that match the specified
	  *  location pattern.
	  *  \brief Look for frame files.
	  *  \param location Location template string.
	  *  \param prefix   Frame file prefix.
	  *  \param vec      frame_loc vector into which the resulting 
	  *                  locators are placed.
	  *  \param drive    Driver type index.
	  */
	void scan(const std::string& location, const std::string& prefix,
		  loc_list& vec, frame_loc::driver_type drive) const;

	/**  Set the debug printout level.
	  *  \brief Set debug level.
	  *  \param lvl Debug printout level.
	  */
	void set_debug(int lvl) ;

	///**  Unlock the location list containing the specified frame_loc.
	//  *  \brief Unlock a list.
	//  *  \param fl frame_loc entry causing the list to be locked.
	//  *  \exception std::runtime_error thrown if the specified 
	//  *             frame_loc can not be found in a list.
	//  */
	//void unlock(const frame_loc* fl) const;

    private:
	/**  Dump the contents of the specified frame map.
	  *  @brief Dump the lists.
	  *  @param frm Source map.
	  *  @param out Output stream.
	  *  @param lvl Verbosity level.
	  *  @return Reference to the output stream.
	  */
	std::ostream& dump_list(const frame_map& frm, std::ostream& out, 
				int lvl) const;

	/**  Return a reference to the map for the specified archive level.
	  *  \brief Reference a specified map.
	  *  \param arch Archive level  of map to be referenced.
	  *  \return Reference to the specified map.
	  */
	frame_map& ref_map(arch_type arch);

	/**  Return a constant reference to the map for the specified 
	  *  archive level.
	  *  \brief Reference a specified map.
	  *  \param arch Archive level  of map to be referenced.
	  *  \return Reference to the specified map.
	  */
	const frame_map& ref_map(arch_type arch) const;

	///**  Return a pointer to the loc_list matching the archive 
	//  *  level and frame type of the specified frame_loc. If the 
	//  *  specified list is not found, a NULL pointer is returned.
	//  *  \brief Reference a specified location list.
	//  *  \param arch Archive level of list to be referenced.
	//  *  \param pfx  Frame type (prefix) of list to be referenced.
	//  *  \return Pointer to the specified location list or NULL.
	//  */
	//loc_list* list_ptr(const frame_loc* floc);

    private:
	frame_map mOnlineMap;
	frame_map mCacheMap;
	frame_map mArchiveMap;
	frame_map mOfflineMap;
	int       mDebug;
    };

    //======================================  Inline methods.
    inline bool
    frame_source::gps_valid(const std::string& type, gps_type gps) const {
	return (list_ptr(type, gps) != NULL);
    }

    //==================================  External frame source.
    extern frame_source Source_List;

}  // namespace sends

#endif  // !defined(SENDS_FRAME_SOURCE_HH)
