/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_LOC_LIST_HH
#define SENDS_LOC_LIST_HH

#include "frame_loc.hh"
#include "epoch.hh"
#include <sortedvector.hh>
#include <gmutex.hh>

// #define LOC_LIST_LOCKED_FINDS 1

namespace sends {

    /**  The %loc_list is a sorted list of frame location (frame_loc) 
      *  segments protected by a read/write mutex.
      *  \brief List of frame location segments.
      *  \author John Zweizig (john.zweizig@ligo.org)
      *  \version $Id$
      */
    class loc_list : public sorted_vector<frame_loc> {
    public:
	/**  Frame data archive level.
	 */
	enum arch_type {
	    /// Online data source
	    kOnline,

	    ///  Cached data (on spinning media)
	    kCache,

	    ///  Archived data (likely on tape)
	    kArchive,

	    ///  Data that can be accessed from an offline source.
	    kOffline
	};

    public:
	/**  Base vector type
	 */
	typedef sorted_vector<frame_loc>  vect_type;
	
	/**  Iterator of base vector type
	 */
	typedef vect_type::iterator       iterator;
 
	/**  Constant iterator of base vector type
	 */
	typedef vect_type::const_iterator const_iterator;

    public:
	/**  Construct an empty locked list, initialize the mutex.
	  *  \brief Default constructor.
	  */
	loc_list(void);

	/**  Destroy the locked list and release the mutex.
	  *  \brief Destructor.
	  */
	virtual ~loc_list(void);

	/**  Get the data archive level for this frame location.
	  *  @brief  Get the archive level.
	  *  @return Archive level.
	  */
	arch_type archive(void) const;

	/**  Get the standard name string for the specified archive level.
	  *  \brief  Archive level name string.
	  *  \param a Archive level to be named.
	  *  \return Archive level name string.
	  */
	static const char* arch_name(arch_type a);

	/**  Coalesce all the segments where possible.
	  *  \brief Coalesce segments.
	  */
	void coalesce(void);

	/**  Copy the contents of the last (most recent) frame locater in 
	  *  the list.
 	  *  \brief Find the specified locater.
	  *  \param fl Instance into which the back entry will be copied.
	  *  \return True is the list is not empty.
	  */
	bool find_back(frame_loc& fl) const;

	/**  Copy the contents of the first frame locater in the list.
 	  *  \brief Find the specified locater.
	  *  \param fl Instance into which the front entry will be copied.
	  *  \return True if the list is not empty.
	  */
	bool find_front(frame_loc& fl) const;

#ifdef LOC_LIST_LOCKED_FINDS
	/**  Find a frame locater covering the specified gps time. The list 
	  *  is read-locked if a frame locater is found. If the gps is not
	  *  included in the list, find_loc returns a NULL pointer and the
	  *  list is left unlocked.
 	  *  \brief Find the specified locater.
	  *  \param gps Time to be located.
	  *  \return pointer to frame locater for the specified gps, or NULL.
	  */
	const frame_loc* find_loc(gps_type gps) const;

	/**  Find a frame locater covering the specified gps time. The list 
	  *  is write-locked if a frame locater is found. If the gps is not
	  *  included in the list, find_loc returns a NULL pointer and the
	  *  list is left unlocked.
 	  *  \brief Find the specified locater.
	  *  \param gps Time to be located.
	  *  \return pointer to frame locater for the specified gps, or NULL.
	  */
	frame_loc* find_loc(gps_type gps);
#endif
	
	/**  Find a frame locater covering the specified gps time and copy its
	  *  contents to the frame_loc referenced by \a fl. If the gps is not
	  *  included in the list, find_loc returns false.
 	  *  \brief Find the specified locater.
	  *  \param gps Time to be located.
	  *  \return True is the locater was found.
	  */
	bool find_loc(gps_type gps, frame_loc& fl) const;

	/**  Convert the archive type name to the corresponding argument 
	  *  type code.
	  *  @brief Convert an archive type string.
	  *  @param name Archive type name.
	  *  @return Archive type code corresponding to the archive type name.
	  */
	static arch_type get_arch_type(const std::string& name);

	/**  Test whether location data exists for the specified gps time.
	  *  \brief Test gps.
	  *  \param gps Time to be tested.
	  *  \return true if data exists.
	  */
	bool in_range(gps_type gps) const;

	/**  Get the memory (in bytes) used by this list.
	  *  \brief memory used by this list.
	  *  \return Number of bytes of memory used by this list.
	  */
	index_type mem_use(void) const;

	/**  Range returns the smallest range containing all available data
	 */
	epoch range(void) const;
	
	/**  Lock the list for read access.
	  *  \brief Read lock.
	  */
	void read_lock(void) const;

	/**  Get a reference to the read/write semaphore for this list.
	  *  This method can be used with semlock to automatically 
	  *  unlock the list when the accessing function terminates.
	  *  \brief Reference read/write mutex.
	  *  \return Reference to the read/write mutex.
	  */
	thread::abstractsemaphore& refmutex(void) const;

	/**  The scan path is used for automatic rescans made by a
	  *  source updater. It is set by the scan/rescan method.
	  *  \brief Path for frame-file scans.
	  *  \return Scan path string.
	  */
	const std::string& scan_path(void) const;

	/**  The scan_time contains the time of the most recent (re)scan 
	  *  for frames of the type contained in this list.
	  *  \brief Most recent scan time.
	  *  \return Most recent scan GPS time.
	  */
	gps_type scan_time(void) const;

	/**  Set the archive level for this location.
	  *  \brief Set the archive level
	  *  \param a Archive type for this location.
	  */
	void set_archive(arch_type a);

	/**  Set the scan path used for automatic rescans.
	  *  \brief Set path for frame-file scans.
	  *  \param path Scan path string.
	  */
	void set_scan_path(const std::string& path);

	/**  Set the time of the most recent (re)scan for frames of the 
	  *  type contained in this list.
	  *  \brief Most recent scan time.
	  *  \param gps Most recent scan GPS time.
	  */
	void set_scan_time(gps_type gps);

	/**  Swap the data vector of the current instance with that of 
	  *  the argument list. Both lists are write-locked before the 
	  *  swap is effectuated. If the the argument list is the same 
	  *  as the current list, no action is taken.
	  *  \note swap will dead-lock if the calling thread owns read 
	  *        or write access to either list.
	  *  \brief Swap data between two lists.
	  *  \param l List containing data to be swapped with the contents
	  *           of this instance.
	  */
	void swap_vec(loc_list& l);

	/**  Unlock the list for write/read access.
	  *  \brief Unlock mutex.
	  */
	void unlock(void) const;

	/**  The upper bound is the lowest entry for which the stop time is 
	  *  greater than the specified gps.
	  *  \brief return upper bound index.
	  *  \param gps Start time searched for.
	  *  \return Index of upper bound locatiuon entry.
	  */
	index_type upper_bound(gps_type gps) const;

	/**  Lock the list for write access.
	  *  \brief Write lock.
	  */
	void write_lock(void);
    private:
	arch_type   _arch_type;
	std::string _scan_path;
	gps_type    _scan_time;
	mutable thread::readwritelock _mutex;
    };

    //==================================  Inline methods.
    inline loc_list::arch_type 
    loc_list::archive(void) const {
	return _arch_type;
    }

    inline thread::abstractsemaphore&
    loc_list::refmutex(void) const {
	return _mutex;
    };

    inline const std::string&
    loc_list::scan_path(void) const {
	return _scan_path;
    }

    inline gps_type
    loc_list::scan_time(void) const {
	return _scan_time;
    }

} // namespace sends

#endif  // !defined(SENDS_FRAME_LOC_HH)
