/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_LOOKUP_STAGE_HH
#define SENDS_LOOKUP_STAGE_HH

#include "types.hh"
#include "seda_stage.hh"
#include <string>
#include <vector>

namespace sends {


    /**  The lookup_stage class provides a minimal template for a SEDA 
      *  stage thread pool. It is based on the seda_stage class and provides
      *  a stage with no filter processing. The constructor, destructor and 
      *  filter methods are implemented. The filter method immediately passes 
      *  the transactions to its output queue (the input queue of the next 
      *  stage in the pipeline.
      *  @brief Channel look-up SEDA stage.
      *  @author John Zweizig
      *  @version 1.0; Last modified December 28, 2007
      *  @ingroup sends_stage
      */
    class lookup_stage : public seda_stage {
    public:
	/**  Source list node (name)
	 */
	typedef std::string source_node;

	/**  Source list datatype.
	 */
	typedef std::vector<source_node> source_list;

    public:
	/**  Construct a lookup_stage object. The arguments indicate the 
	  *  number of threads in the pool and the number of input transactions
	  *  that may be queue simultaneously (note that the pipline gets 
	  *  indigestion only after more than (nTrans + nThread) trasactions
	  *  need to execute in this stage.
	  *  @brief Stage template constructor.
	  *  @param nThread Number of threads to run simultaneously in the pool.
	  *  @param nTrans  Number of transactions that may be queued.
	  */
	lookup_stage(int nThread, int nTrans);

	/**  Stage template destructor is null by default. It is mostly
	  *  useful to remove resources allocated by the stages.
	  */
	virtual ~lookup_stage(void);

	/**  Build a list of channel data source names.
	  *  @brief Build list of data sources.
	  *  @param start Start time to be checked. 
	  *  @param trans Pointer to the current transaction 
	  *  @param svec  Vector of source names to be filled.
	  */
	void build_source_list(gps_type start, transaction* trans, 
			       source_list& svec);

	/**  The filter method performs the transaction processing for a 
	  *  single stage. By default this is a null filter that passes 
	  *  the transaction on to the next 
	  *  @brief Transaction filter for this stage.
	  *  @param trans Pointer to the current transaction 
	  */
	void filter(transaction* trans);

	/**  Return a constant pointer to a null-terminated character vector 
	  *  containing the name of the stage.
	  *  @brief Stage name.
	  *  @return Constant pointer to the stage name string.
	  */
	const char* stage_name(void) const;
    private:
    };
}

#endif // !defined(SENDS_LOOKUP_STAGE_HH)
