/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_LSMP_DRIVER_HH
#define SENDS_LSMP_DRIVER_HH

#include "frame_driver.hh"
#include "gmutex.hh"
#include <iosfwd>

class iSMbuf;

namespace FrameCPP {
    namespace Common {
	class FrameBufferInterface;
    }
}

namespace sends {
    class loc_list;

    /**  NDS2 frame input driver for frames residing in a LIGO shared memory
      *  partition.
      *  \brief LSMP frame input driver.
      *  \ingroup sends_drive
      *  \version 1.0; last modified August 7, 2008.
      *  \author John Zweizig
      */
    class lsmp_driver : public frame_driver {
    public:

	/**  Construct a shared memory driver.
	  *  \brief Constructor
	  */
	lsmp_driver(void);

	/**  Construct a shared memory driver with a specified partition.
	  *  \brief Constructor
	  *  \param part Partition name.
	  */
	lsmp_driver(const std::string& part);

	/**  Destroy a share memory driver.
	  *  \brief Destructor
	  */
	~lsmp_driver(void);

	/**  Get the driver specific name ("lsmp:<partition>").
	  *  \brief get the driver name.
	  *  \return driver name ("lsmp:<partition>").
	  */
	std::string name(void) const;

	/**  Get the gps time of currently available online data. If no data 
	  *  segment is available a new segment is read and the time is 
	  *  returned.
	  *  \brief Current online data time.
	  *  \return Start time of online data.
	  */
	virtual gps_type online_gps(void);

	/**  Test whether this is an online device.
	  *  \brief test for online device.
	  *  \return true if devices is capable of online I/O.
	  */
	virtual bool online(void) const;

	/**  Get the driver partition name.
	  *  \brief get the partition name.
	  *  \return parition name.
	  */
	const char* part(void) const;

	/**  Read data from the specified path for the specified times. 
	  *  Channel names and segments they are to be read into are passed
	  *  as a list.
	  *  \brief Read channel ldata into segments. 
	  *  \param type  Source frame location list.
	  *  \param path  Source descriptor path field.
	  *  \param start Start gps time of data to be read.
	  *  \param end   End gps of data to be read.
	  *  \param lst   List of channels and segment IDs.
	  *  \return Result code. 0 if all requested data were read,
	  *          otherwise -1.
	  */
	virtual int read_data(const loc_list& type, gps_type start, 
			      gps_type dt, const channel_vect& lst);

	/**  Test whether data are available from this device.
	  *  \brief Test for available data.
	  *  \return true if device has online data.
	  */
	virtual bool ready(void) const;

	/**  Decrement the use count for the specified channel in the 
	  *  driver's requested channel list. remove the channel if the
	  *  use count drops to zero.
	  *  \brief Configure to request the specified channel.
	  *  \param c   Index of channel to be removed from the request list.
	  *  \return Result code. 0 if channel release succeded, otherwise -1.
	  */
	virtual int release_channel(chan_index c);

	/**  Add the specified channel to the driver's request list.
	  *  \brief Configure to request the specified channel.
	  *  \param c   Index of channnel to add to the request list.
	  *  \return Result code. 0 if channel request succeded, otherwise -1.
	  */
	virtual int request_channel(chan_index c);


	/**  test whether data of the specified type are available in the 
	  *  specified path for the specified times.
	  *  \brief Read channel ldata into segments.
	  *  \param type Source frame type field.
	  *  \param path Source descriptor path field.
	  *  \param t    Start gps time of data to be read.
	  *  \return True if the specified data frame exists.
	  */
	virtual bool test_available(const loc_list& ll, gps_type t) const;

    private:
	struct chan_node {
	    chan_node(chan_index chan);
	    chan_index _chan;
	    long       _count;
	};
	typedef std::vector<chan_node> chan_list;


    private:
	/**  Get the index of the first element containing a channel index 
	  *  greater than or equal to the specified channel index.
	  *  \return Location or insertion point for the specfied channel.
	  */
	index_type chanlist_index(chan_index c) const;

	/**  Dump the contents of the channel list to the specified stream.
	  *  \brief Dump channel list.
	  *  \param output stream
	  */
	void dump_chanlist(std::ostream& out) const;

	/**  Find the channel list entry for the specified channel.
	  *  @brief  Find channel entry.
	  *  @param name Channel name.
	  *  @return Reference to channel entry.
	  */
	chan_node& find_channel(chan_index c);

	/**  Find the channel list entry for the specified channel.
	  *  @brief  Find channel entry.
	  *  @param name Channel name.
	  *  @return Constant reference to channel entry.
	  */
	const chan_node& find_channel(chan_index c) const;

	/**  Insert a channel list entry for the specified channel. The 
	  *  use count is set to 0. If the channel is already in the list, 
	  *  the entry is 
	  *  @brief  Find channel entry.
	  *  @param name Channel name.
	  *  @return Reference to channel entry.
	  */
	chan_node& insert_channel(chan_index c);

	/**  Start the lsmp consumer for the specified partition.
	  *  @brief Connect an input stream to the partition.
	  *  @return true if successful.
	  */
	bool start_consumer(void);

	/**  Close the lsmp consumer for the specified partition.
	  *  @brief Connect an input stream to the partition.
	  *  @return true if successful.
	  */
	bool stop_consumer(void);

    private:
	enum state_bits {
	    kStateOffline, // partition not available
	    kStateError,   // errror fetchin data from partition
	    kStateTimeOut  // data not arriving in partition
	};
	void set_state(state_bits bit, bool val=true);
	bool test_state(state_bits bit) const;

    private:
	thread::mutex mGate;
	thread::readwritelock mChanMutex;

	std::string  mPartition;
	iSMbuf*      mLSMPbuffer;
	FrameCPP::Common::FrameBufferInterface* mFrameBuf;
	chan_list  mChanList;
	gps_type   mActiveChan;
	gps_type   mNextGPS;
	mask_type  mState;
    };

    //==================================  Inline methods
    inline
    lsmp_driver::chan_node::chan_node(chan_index chan)
	: _chan(chan), _count(0)
    {}

    inline bool
    lsmp_driver::online(void) const {
	return true;
    }

    inline const char* 
    lsmp_driver::part(void) const {
	return mPartition.c_str();
    }

    inline bool
    lsmp_driver::ready(void) const {
	return !mState;
    }

    inline bool
    lsmp_driver::test_state(state_bits bit) const {
	return (mState & (1 << bit)) != 0;
    }

}  // namspace sends

#endif   // !defined(SENDS_FILE_DRIVER_HH)
