/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_PARSE_STAGE_HH
#define SENDS_PARSE_STAGE_HH

#include "seda_stage.hh"
#include "buffer_pool.hh"

namespace sends {

    const int parse_buffer_length=32768;
    class buffer;

    /**  The parse_stage class is the command parser stage of the SENDS 
      *  server. For each transaction entered into the input queue, the
      *  parse stage reads a command line (terminated by a new-line or null)
      *  and parses the command. The command format is:
      *
      *  \verbatim
      <command> [<start> [<end>] <delta>] [<type>] [{<chan1> [... <chanN>]}];
         \endverbatim
      *
      *  The command may be one of the following:
      *  - noop: The command is parsed but no data are transferred. 
      *  - get-channels: Get channel data for specified time and types.
      *  - get-data: Get specified data.
      *  - get-status: Get status of specified channels.
      *  - get-server-status: Get list of server info.
      *  - quit: Close the client connection.
      *
      *  The data times are specified by a start GPS time and either an
      *  end time, a data stride or both. If two time fields are specified 
      *  and the second is less than the first the second field is taken
      *  to be a duration (delta) time and the end time is caculated from
      *  the start minus the end. If the second field is greater than the 
      *  first, or three fields are specified, the second field specifies
      *  the end time and the optional third field specifies the data stride
      *  (segment length for a single transfer).
      *
      *  The channel list contains an arbitrary number (up to 1024) of space 
      *  separated channel descriptors enclosed in braces. Each channel 
      *  descriptor has a channel name and optional channel-type, sample-rate
      *  or both separated by commas or percent signs.
      *
      *  @brief Command parser SEDA stage.
      *  @author J. Zweizig
      *  @version 1.0; Last modified December 28, 2007
      *  @ingroup sends_stage
      */
    class parse_stage : public seda_stage {
    public:
	/**  Data type for preallocated buffer free list.
	  *  @brief Buffer list.
	  */
        typedef thread::mt_queue<buffer*> buffer_queue;
        // typedef tbb::concurrent_queue<buffer*> buffer_queue;
    public:
	/**  Construct a parse_stage object. The arguments indicate the 
	  *  number of threads in the pool and the number of input transactions
	  *  that may be queue simultaneously (note that the pipline gets 
	  *  indigestion only after more than (nTrans + nThread) trasactions
	  *  need to execute in this stage.
	  *  @brief Stage template constructor.
	  *  @param nThread Number of threads to run simultaneously in the pool.
	  *  @param nTrans  Number of transactions that may be queued.
	  */
	parse_stage(int nThread, int nTrans);

	/**  Stage template destructor is null by default. It is mostly
	  *  useful to remove resources allocated by the stages.
	  */
	virtual ~parse_stage(void);

	/**  The filter method performs the transaction processing for a 
	  *  single stage. By default this is a null filter that passes 
	  *  the transaction on to the next 
	  *  @brief Transaction filter for this stage.
	  *  @param trans Pointer to the current transaction 
	  */
	void filter(transaction* trans);

	/**  Set up the signal handling for this thread.
	  *  \brief Thread initialization.
	  */
	void init_thread(void);

	/**  Reference the resource (input buffer) queue.
	  */
        buffer_pool& ref_resource(void);

	/**  Set the parameters for this stage.
	  *  @brief Set stage parameters.
	  *  @param p parameter block.
	  */
	void set_parameters(const stage_pars& p);

	/**  Shut down all threads in the parse pool.
	  *  \brief Shut down all threads in the pool.
	  */
	virtual void shut(void);

	/**  Return a constant pointer to a null-terminated character vector 
	  *  containing the name of the stage.
	  *  @brief Stage name.
	  *  @return Constant pointer to the stage name string.
	  */
	const char* stage_name(void) const;

    private:
	void parse_channels(const char* &ptr, const char* end, transaction* t);
	void parse_chantype(const char* &ptr, const char* end, transaction* t);
	void parse_command(const char* &ptr, const char* end, transaction* t);
	void parse_epoch(const char* &ptr, const char* end, transaction* t);
	void parse_pattern(const char* &ptr, const char* end, transaction* t);
	int  parse_times(const char* &ptr, const char* end, int nMax, 
			 long iarg[]);

    private:
        buffer_pool resource;
	double      timeout;

	//------------------------------  Statistics
	long        quit_count;
    };

    inline buffer_pool&
    parse_stage::ref_resource(void) {
	return resource;
    }
}

#endif // !defined(SENDS_PARSE_STAGE_HH)
