/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_SEDA_STAGE_HH
#define SENDS_SEDA_STAGE_HH

#include "thread_pool.hh"
#include "transaction.hh"
#include "gmutex.hh"

namespace sends {

    /**  The Seda Stage class provides the base for a server pipeline stage.
      *  each pipline stage uses a pool of threads to perform one stage of
      *  processing in the pipeline. The events to be processed are 
      *  represented by tokens (transaction objects) placed in the input 
      *  queue. After process by each stage  is completed the tokens are 
      *  passed on to the next stage (or alternate stage) as indicated by 
      *  the output and alternate .
      *  @brief Seda pipeline stage base class
      *  @author John Zweizig
      *  @version 1.0; Last modified April 25, 2008
      *  @ingroup sends_stage
      */
    class seda_stage : public thread_pool {
    public:
	/**  Debug level ennumeration.
	  */
	enum dbg_level {
	    dlvl_none=0,       ///< No debugging output
	    dlvl_transaction,  ///< Print debug message per transaction.
	    dlvl_meta,         ///< Print substage messages
	    dlvl_detail,       ///< Print detailed control-flow messages
	    dlvl_data          ///< Print data
	};

	/**  Error level ennumeration.
	  */
	enum err_level {
	    elvl_log=0,   ///< Log message
	    elvl_warning, ///< Warning message
	    elvl_error,   ///< Error message
	    elvl_severe   ///< Severe error message
	};

	/**  Stage parameter structure for setting stage parameters.
	  *  \brief Stage parameter structure
	  */
	struct stage_pars {
	    ///  Debug printout level.
	    dbg_level _debug;
	    ///  Input queue capacity.
	    int _inq_capacity;
	    ///  Maximum threads.
	    int _threads;
	    ///  Maximum wait time
	    double _timeout;
	};

    public:
	/**  Construct a stage with the specified number of threads in 
	  *  the thread pool and the specified number of transactions in
	  *  the input queue.
	  *  @brief Stage constructor.
	  *  @param nThread Initial number of threads in pool.
	  *  @param nTrans  Number of transation sots in input queue.
	  */
	seda_stage(int nThread, int nTrans=16);

	/**  Stage destructor.
	  *  @brief Destructor.
	  */
	virtual ~seda_stage(void);

	/**  Reset a transaction and close the connection. Log the
	  *  descriptive message.
	  *  @brief Cancel a transaction.
	  *  @param trans Pointer to transaction object.
	  *  @param msg   Message indicating reason for the cancellation.
	  */
	void cancel(transaction* trans, const std::string& msg);

	/**  Return debug print level for this stage.
	  *  @brief Debug level.
	  *  @return Debug print level.
	  */
	dbg_level debug(void) const;

	/**  Test if debug level is greater than the specified value.
	  *  @brief Test debug level.
	  *  @param lvlmin Minimum debug level.
	  *  @return True if debug print level is greater than threshold.
	  */
	bool debug(dbg_level lvlmin) const;

	/**  Print the specified message if appropriate for the debug level.
	  *  @brief Conditionally print message.
	  *  @param lvlmin Minimum debug level.
	  *  @param id     Transacrtion ID.
	  *  @param msg    Message to print out
	  */
	void debug_msg(dbg_level lvlmin, long id, const std::string& msg) const;

	/**  Print the specified message, formatted as appropriate for the 
	  *  error level.
	  *  @brief Conditionally print error message.
	  *  @param rlvl Error level.
	  *  @param t    Transaction descriptor.
	  *  @param msg  Message to print out
	  */
	void error_msg(err_level rlvl, const transaction& t, 
		       const std::string& msg) const;

	/**  Method to perform pipeline stage processing on the transaction
	  *  passed as an argument.
	  *  @brief Stage processing method.
	  *  @param trans Trnsaction to be processed by this method.
	  */
	virtual void filter(transaction* trans) = 0;

	/**  Method to perform initialization processing for each thread.
	  *  @brief Thread initialization method.
	  */
	virtual void init_thread(void);

	/**  Link the alternate queue pointer to the input of the specified 
	  *  stage.
	  *  @brief Link atlernate output to the next stage.
	  *  @param pool Next stage in the pipeline.
	  */
	void link_alternate(seda_stage& pool);

	/**  Link the output queue pointer to the input of the specified 
	  *  stage.
	  *  @brief Link output to the next stage.
	  *  @param pool Next stage in the pipeline.
	  */
	void link_error(seda_stage& pool);

	/**  Link the output queue pointer to the input of the specified 
	  *  stage.
	  *  @brief Link output to the next stage.
	  *  @param pool Next stage in the pipeline.
	  */
	void link_output(seda_stage& pool);

	/**  Enter the transaction into the alternate queue (i.e. the input 
	  *  queue of the branch stage in the pipeline).
	  *  @brief Pass transaction to the branch pipeline stage.
	  *  @param trans Transaction to be passed to the alternate output.
	  */
	void put_alternate(transaction* trans);

	/**  Free transaction resources and place trasaction into the error 
	  *  queue (generally, the input queue of the command-wait stage.)
	  *  If the error queue has not been defined, the transaction is 
	  *  released.
	  *  @brief Pass transaction to the error stage.
	  *  @param trans Transaction to be passed to the error state.
	  */
	void put_error(transaction* trans);

	/**  Enter the transaction into the output queue (i.e. the input 
	  *  queue of the next stage in the pipeline.
	  *  @brief Pass transaction to the next pipeline stage.
	  *  @param trans Transaction to be passed to the output.
	  */
	void put_output(transaction* trans);

	/**  Return a constant reference to the input queue for this stage.
	  *  @brief Reference the input queue.
	  *  @return Reference to the input queue.
	  */
	const tran_queue& ref_input(void) const;

	/**  Return a reference to the input queue for this stage.
	  *  @brief Reference the input queue.
	  *  @return Reference to the input queue.
	  */
	tran_queue& ref_input(void);

	/**  Reset a transaction and place it in the error queue. If no 
	  *  error queue is defined, the transaction is trashed. Log the
	  *  descriptive message.
	  *  @brief Reject a transaction.
	  *  @param trans Pointer to transaction object.
	  *  @param msg   Message indicating reason for the rejection.
	  */
	void reject(transaction* trans, const std::string& msg);

	/**  Set the debug print level for this stage.
	  *  @brief Set debug level.
	  *  @param lvl value of debug print level for this stage.
	  */
	void set_debug(dbg_level lvl);

	/**  Set the parameters for this stage.
	  *  @brief Set stage parameters.
	  *  @param p parameter block.
	  */
	virtual void set_parameters(const stage_pars& p);

	/**  Shut down all threads in the pool by doing the following:
	  *    - Set the pool size to zero. This will prevent the pool 
	  *      from starting new threads and kill threads as they are
	  *      released.
	  *    - Shut down the input queue. This will release any waiting 
	  *      threads and prevent any further transactions from being 
	  *      added to the queue.
	  *    - release any transactions remaining in the pool.
	  *    - print statistics
	  *  \brief Shut down all threads in the pool.
	  */
	virtual void shut(void);

	/**  Return a constant pointer to the character vector containing 
	  *  the name of this stage.
	  *  @brief Stage name.
	  *  @return pointer to the stage name.
	  */
	virtual const char* stage_name(void) const;

	/**  Perform pipeline stage startup processing including starting 
	  *  up the initial number of threads.
	  *  @brief Start pipeline stage.
	  */
	void start(void);

	/**  Print the specified message, formatted as appropriate for the 
	  *  error level. A system error message corresponding to the current
	  *  error number (errno) is appended to the message.
	  *  @brief Conditionally print message.
	  *  @param rlvl Error level.
	  *  @param t    Transaction descriptor.
	  *  @param msg  Message to print out
	  */
	void systm_msg(err_level rlvl, const transaction& t, 
		       const std::string& msg) const;

	/**  Common stage code that provides synchronization and account
	  *  between thread in the same stage.
	  *  @brief Thread entry point.
	  */
	virtual void thread_entry(void);

	/**  Convert a debug level to its corresponding name string.
	  *  @param code Debug level code to be converted.
	  *  @return Debug level name string.
	  */
	static const char* dbg_lvl_name(dbg_level code);

	/**  Convert a debug level name string to a dbg_level code.
	  *  @param  name Debug level name string.
	  *  @return Converted debug level code.
	  */
	static dbg_level   dbg_lvl_code(const std::string& name);
	
    private:
	/**  Mutex for serializing debug message printing.
	 */
	static thread::mutex msg_mutex;

    private:
	dbg_level    dbglvl;
	tran_queue   input;
	tran_queue*  output;
	tran_queue*  alternate;
	tran_queue*  error_queue;
	long         primary_count;
	long         alternate_count;
	long         reject_count;
    };

    //==================================  Inline methods
    inline seda_stage::dbg_level
    seda_stage::debug(void) const {
	return dbglvl;
    }

    inline bool
    seda_stage::debug(dbg_level minlvl) const {
	return dbglvl >= minlvl;
    }

    inline tran_queue&
    seda_stage::ref_input(void) {
	return input;
    }

    inline const tran_queue&
    seda_stage::ref_input(void) const {
	return input;
    }

}

#endif // !defined(SENDS_SEDA_STAGE_HH)
