/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_SEG_POOL_MGR
#define SENDS_SEG_POOL_MGR

#include "pool_mgr.hh"
#include "buffer_pool.hh"
#include <vector>

namespace sends {

    /**  Class to manage segment buffer pools of different lengths.
      *  @brief Buffer pool manager API.
      *  @author John Zweizig
      *  @version 1.0;Last modified February 12, 2008
      *  @ingroup sends_base
      */
    class seg_pool_mgr : public pool_mgr {
    public:
	/**  Type of data used to select a buffer class.
	  */
        using pool_mgr::select_type;
    
	/**  Type of data use to represent a buffer class.
	  */
        using pool_mgr::bufclass_type;

	/**  Buffer type enumerator.
	  */
	enum buffer_class_id {
	    kSmallBuf,  ///< Small buffer class
	    kLargeBuf,  ///< Large buffer class
	    kXLargeBuf  ///< Extra-large buffer class
	};

    public:
	/**  Construct an empty pool manager.
	  *  @brief Null constructor.
	  */
	seg_pool_mgr(void);

	/**  Destroy a pool manager and deallocate the managed buffer pools.
	  *  @brief Pool manager destructor.
	  */
	~seg_pool_mgr(void);

	/**  Create a buffer pool with nbuf buffers of length lbuf. Add the 
	  *  pool  to the manager with the specified id.
	  *  @brief Add a buffer pool.
	  *  @param id Pool id.
	  *  @param nbuf Number of buffer in pool.
	  *  @param lbuf Length of buffer in pool
	  */
	void add_pool(bufclass_type id, long nbuf, long lbuf);

	/**  Get the number of buffers allocated in the pool for the 
	  *  specified class.
	  *  @brief Get buffer count.
	  *  @return Number of buffers in the specified pool.
	  */
	index_type buffer_count(bufclass_type bufclass) const;

	/**  Decide on a buffer class based on the selection code.
	  *  @brief Chose a buffer class.
	  *  @param selid Selection code.
	  *  @return Buffer class for selid.
	  */
	virtual bufclass_type get_buffer_class(select_type selid) const;

	/**  Allocate a buffer of the specified class.
	  *  @brief Chose a buffer class.
	  *  @param selclass buffer class code.
	  *  @return Pointer to requested buffer.
	  */
	virtual buffer* get_buffer(select_type selclass);

	/**  Allocate a buffer of the specified class without waiting.
	  *  @brief Chose a buffer class.
	  *  @param selclass buffer class code.
	  *  @return Pointer to requested buffer.
	  */
	virtual buffer* get_buffer_now(select_type selclass);	

	/**  Get the size of the buffers in the specified class.
	  *  @brief Get buffer size
	  */
	index_type size(bufclass_type bufclass) const;

    private:
	typedef std::vector<buffer_pool*> class_vect;
	class_vect mPools;
    };

    //==================================  Inline methods
    inline index_type
    seg_pool_mgr::buffer_count(bufclass_type ibc) const {
	if (ibc > mPools.size() || !mPools[ibc]) return 0;
	return mPools[ibc]->size();
    }

    inline index_type
    seg_pool_mgr::size(bufclass_type ibc) const {
	if (ibc > mPools.size() || !mPools[ibc]) return 0;
	return mPools[ibc]->buffer_size();
    }

    //==================================  External Buffer_Pool
    extern seg_pool_mgr Buffer_Pool;



}  // namespace sends

#endif // !defined(SENDS_POOL_MGR)
