/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_SEGDB_HH
#define SENDS_SEGDB_HH

#include "segment.hh"
#include "gmutex.hh"
#include <iosfwd>
#include <vector>
#include <memory>

namespace sends {

    class pool_mgr;

    /**  The %segment database class contains a list of all memory-resident 
      *  channel data segments. Each data-segment is represented by a %segment
      *  object in the the pre-allocated vector. The segment objects are 
      *  linked in two doubly-linked lists:
      *   - a list of all segments in least recently used order. 
      *   - a list per channel of segments containing data from that channel.
      *     These lists are ordered according to the data start times.
      *
      *  The %linkage information for the two lists is maintained in separate 
      *  vectors in the database. Since each %segment is in a unique channel
      *  list, all channel lists use the same %linkage list.
      *
      *  When data from a specific channel and gps time are requested, the
      *  data manager client request a %segment number for the specified 
      *  channel and time. This is accomplished by:
      *  \verbatim
         seg_index iseg = segdb.find(chanId, start, stop); \endverbatim
      *
      *	 or,
      *  \verbatim
	 seg_index iseg = segdb.reserve(chanId, gps);  \endverbatim
      * 
      *  The reserve() method differs from find() in that it increments the 
      *  %segment use-count and moves it to the end of the age list. If 
      *  either method is unable to find the specified data it returns the 
      *  end index which can be tested with is_null().
      *
      *  @brief Segment database class
      *  @author J. Zweizig
      *  @version 1.1; Last Modified January 14, 2008
      *  @ingroup sends_base
      */
    class seg_db {
    public:
	/**  Segment vector data type.
	 */
	typedef std::vector<segment> seg_vect;

	/**  Segment linkage data contains a list with a foreward, and a 
	  *  back pointer for each segment. The "null" pointer is assigned 
	  *  to the backward pointer of the first element and foreward 
	  *  pointer of the last element of each list.
	  *  @brief Segment linkage class.
	  *  @ingroup sends_base
	  */
	class linkage {
	public:
	    /**  Construct a linkage list with nSegs entries.
	      *  @brief Construct a linkage vector.
	      *  @param nSegs Initial number of entries for the list.
	      */
	    linkage(seg_index nSegs);

	    /**  Get the end segment index. This is used as the null pointer.
	      *  @brief Get the end segment index.
	      *  @return Index of segment before the argument segment.
	      */
	    seg_index end(void) const;

	    /**  Insert segment 'inx' into the linked list before segment
	      *  iaft.
	      *  @brief Insert a segment into the linkage list.
	      *  @param inx  Index of segment to be inserted
	      *  @param iaft Segment in list before which the new segment 
	      *              will be inserted. 
	      */
	    void insert_before(seg_index inx, seg_index iaft);

	    /**  Insert segment 'inx' into the linked list after segment
	      *  past.
	      *  @brief Insert a segment into the linkage list.
	      *  @param inx  Index of segment to be inserted
	      *  @param past Segment in list after which the new segment 
	      *              will be inserted. 
	      */
	    void insert_after(seg_index inx, seg_index past);

	    /**  Get the index of the segment following the argument segment 
	      *  in the linked list.
	      *  @brief Get the next segment index.
	      *  @param inx Current segment index.
	      *  @return Index of segment following the argument segment.
	      */
	    seg_index next(seg_index inx) const;

	    /**  Test whether the specified index is null.
	      *  @brief test if index is null (equal to end()).
	      *  @param inx Index to be tested.
	      *  @return True if index is null.
	      */
	    bool null(seg_index inx) const;

	    /**  Get the index of the segment before the argument segment 
	      *  in the linked list.
	      *  @brief Get the previous segment index.
	      *  @param inx Current segment index.
	      *  @return Index of segment before the argument segment.
	      */
	    seg_index prev(seg_index inx) const;

	    /**  Remove the specified segment from the linked list. The 
	      *  previous and next segments in the list are linked together
	      *  and both links for the specified segment are replaced by 
	      *  the null link value (end).
	      *  @brief Unlink a segment.
	      *  @param inx Index of the segment to be unlinked.
	      */
	    void unlink(seg_index inx);

	private:
	    typedef std::pair<seg_index,seg_index> dll_node;
	    typedef std::vector<dll_node> dll_vect;
	    seg_index mEnd;
	    dll_vect  mVect;
	};

	/**  The queue base class containes first and last pointers to
	  *  anchor a linked list.
	  *  @brief Linked list anchor.
	  *  @ingroup sends_base
	  */
	class qbase {
	public:
	    /**  Construct a queue base using N as the null value.
	      *  @brief Contruct a qbase.
	      *  @param N Null pointer value.
	      */
	    qbase(seg_index N);

	    /**  Destroy the queue base.
	      *  @brief Destructor.
	      */
	    ~qbase(void);

	    /**  Get the index of the first entry in the list.
	      *  @brief Get first entry index.
	      *  @return Index of first entry in the list.
	      */
	    seg_index first(void) const;

	    /**  Insert an entry into the specified linkage list before 
	      *  the specified entry. No write lock is performed - the 
	      *  queue must be locked before calling insert.
	      *  @brief Insert an entry.
	      *  @exception std::logic_error Insert attempted on an 
	      *             unlocked queue.
	      *  @param iseg Index of segment to be inserted.
	      *  @param ib4  Position in the list.
	      *  @param l    Linkage list.
	      */
	    void insert(seg_index iseg, seg_index ib4, linkage& l);

	    /**  Get the index of the last entry in the list.
	      *  @brief Get last entry index.
	      *  @return Index of last entry in the list.
	      */
	    seg_index last(void) const;

	    /**  Get the index of the next entry in the list.
	      *  @brief Get next entry index.
	      *  @param iseg Index of current segment
	      *  @param l    Linkage vector.
	      *  @return Index of last entry in the list.
	      */
	    seg_index next(seg_index iseg, const linkage& l) const;

	    /**  Get the index of the previous entry in the list.
	      *  @brief Get previous entry index.
	      *  @param iseg Index of current segment
	      *  @param l    Linkage vector.
	      *  @return Index of previous entry in the list.
	      */
	    seg_index prev(seg_index iseg, const linkage& l) const;

	    /**  Lock the segment queue for read access.
	      *  @brief Lock for read access.
	      */
	    void read_lock(void) const;

	    /**  Return a reference to the read/write queue access mutex.
	      *  @brief Return reference to the mutex.
	      *  @return Reference to the mutex.
	      */
	    thread::readwritelock& ref_mutex(void) const;

	    /**  Remove the specified entry from the specified list. No write
	      *  lock of the structure  is performed.
	      *  @brief Remove an entry.
	      *  @exception std::logic_error Remove attempted on an 
	      *             unlocked queue.
	      *  @param iseg segment to be removed.
	      *  @param l    Linkage list from which the entry will be removed.
	      */
	    void remove(seg_index iseg, linkage& l);

	    /**  The queue lock is tested for write access. If the queue is
	      *  not locked, an exception is thrown
	      *  @brief Unlock access.
	      *  @exception std::logic_error The queue is not locked for
	      *             write access.
	      */
	    void testlock(void) const;

	    /**  Unlock the segment queue.
	      *  @brief Unlock access.
	      */
	    void unlock(void) const;

	    /**  Lock the segment queue for write access.
	      *  @brief Lock for write access.
	      */
	    void write_lock(void);

	private:
	    seg_index mFirst;
	    seg_index mLast;
	    mutable thread::readwritelock mRWMutex;
	};

	/**  Channel queue anchors.
	  */
	typedef std::vector<qbase> base_vect;

    public:
	/**  Construct a segment database with space for the specified number 
	  *  of segments and number of channels.
	  *  @brief construct a segment database.
	  *  @param nSeg Number of segments to pre-allocate.
	  *  @param nChan Number of channel slots (lists) to pre-allocate.
	  */
	seg_db(seg_index nSeg, chan_index nChan);

	/**  Destroy the segment database
	  *  @brief seg_db destructor
	  */
	~seg_db(void);

	/**  Add an empty %segment with the specified channel and start time.
	  *  If the new segment overlaps unused space in the previous segment,
	  *  the previous %segment is truncated. If the next segment in the 
	  *  channel list overlaps the available space in the new segment, the
	  *  new segment is truncated. If the specified start time overlaps 
	  *  currently available data or reserved storage in an existing 
	  *  segment, the existing segment number is returned. The channel 
	  *  list must be available for writing. 
	  *  \brief Add a %segment to contain the data.
	  *  \param pmgr  Buffer pool manager.
	  *  \param chan  ID of channel for which the buffer with be allocated.
	  *  \param rate  Channel sample rate.
	  *  \param dtyp  Channel data type.
	  *  \param start Start time for which the buffer will be reserved.
	  *  \param mind  Minimum required duration.
	  *  \return Index of the new (or existing) segment.
	  */
	seg_index add_segment(pool_mgr& pmgr, chan_index chan, double rate, 
			      data_type dtyp, gps_type start, gps_type mind);

	/**  Test whether the channel index is valid
	  *  @brief Channel index test
	  *  @param chan Channel index to test
	  *  @return true if index is in valid range.
	  */
	bool chan_valid(chan_index chan) const;

	/**  Check the segment database by looping over all segments
	  *  and verifying the forward and backward links.
	  */
	void check(void) const;

	/**  Clear the segments database by looping over all channel queues
	  *  and releasing the associated segments.
	  */
	void clear(void);

	/**  Dump out the segment list.
	  *  @brief Dump out the segment list
	  *  @param out Output stream reference
	  *  @return Output stream reference
	  */
	std::ostream& dump(std::ostream& out) const;

	/**  Dump the segment list by channel chain.
	  *  @brief Dump the segment list by channel
	  *  @param out Output stream reference
	  *  @return Output stream reference
	  */
	std::ostream& dump_by_channel(std::ostream& out) const;

	/**  Return the segment list null index.
	  *  @brief Null index
	  *  @return Null index.
	  */
	seg_index end(void) const;

	/**  Find the first segment containing the specified times 
	  *  for the specified channel. If no segments are found in 
	  *  the specified channel and times, the null index is 
	  *  returned. find() locks the channel queue while searching 
	  *  for the data.
	  *  @brief Find a segment.
	  *  @param chan Channel index
	  *  @param start Start time.
	  *  @param stop End time.
	  *  @return Index of first segment in the specified range.
	  */
	seg_index find(chan_index chan, gps_type start, gps_type stop) const;

	/**  Find the first segment for the specified channel having a start 
	  *  time greater than the specified gps.
	  *  @brief find a segment.
	  *  @param chan Channel index
	  *  @param gps Compare time.
	  *  @return Index of first segment after the specified time.
	  */
	seg_index first_gt(chan_index chan, gps_type gps) const;

	/**  Insert the specified index into the front (oldest entry) of 
	  *  the age queue.
	  *  @brief Free a segment.
	  *  @param iseg Index of segment to be freed.
	  */
	void free(seg_index iseg);

	/**  Get the position of the specified segment in the age queue. 
	  *  the last (youngest) segment in the queue is number 0. A null
	  *  position is returned if the segment is not found.
	  *  @brief Get the segment's age.
	  *  @param i Segment number
	  *  @return Age of specified segment
	  */
	seg_index get_age(seg_index i) const;

	/**  Get a segment descriptor, that will accomodate, the specified
	  *  data. The reserve pointer is set to the stop time or the end of
	  *  the buffer, as appropriate. The use count of the identified 
	  *  %segment is incremented and the fill lock is set if a valid 
	  *  %segment number is returned. A null %segment pointer (end()) is
	  *  returned if the %segment already contains the data or has space 
	  *  reserved for the requested data.
	  *  @brief Find a %segment to contain the data.
	  *  @param pmgr  Buffer pool manager.
	  *  @param chan  ID of channel for which the buffer with be allocated.
	  *  @param rate  Channel sample rate.
	  *  @param dtyp  Channel data word type.
	  *  @param start Start time for which the buffer will be reserved.
	  *  @param stop  End time for which the data will be reserved.
	  *  @return Segment index.
	  */
	seg_index get_buffer(pool_mgr& pmgr, chan_index chan, double rate, 
			     data_type dtyp, gps_type start, gps_type stop);

	/**  Get a free segment descriptor. If an unused buffer is available, 
	  *  an unused %segment is allocated and the buffer is attached. If 
	  *  no unused buffers or segments are found, the oldest appropriate 
	  *  unused %segment is used. The descriptor is unlinked from both the
	  *  age list and the channel list and it's index is returned.
	  *  @brief Find and unlink a free segment
	  *  @param pmgr Pool manager.
	  *  @param buf_class Buffer class
	  *  @return Index of the free segment.
	  */
	seg_index get_segment(pool_mgr& pmgr, int buf_class);

	/**  Get an unused segment descriptor with the specified buffer class, 
	  *  starting from the least recently used and working through the 
	  *  age list until an unused segment descriptor is found. The 
	  *  descriptor is unlinked from both the age list and the channel 
	  *  list and its index is returned.
	  *  @brief Find and unlink a free segment
	  *  @param buf_class Buffer class
	  *  @return Index of the free segment.
	  */
	seg_index getOldest(int buf_class);

	/**  Insert the specified segment number into the age and channel lists.
	  *  @brief Insert a segment. 
	  *  @param chan Channel ID.
	  *  @param inx  Index of segment to be inserted.
	  */
	void insert(chan_index chan, seg_index inx);

	/**  Test if the specified index is the null pointer.
	  *  @brief Test for null pointer
	  *  @param seg Index to be tested for null value.
	  *  @return True if pointer is null.
	  */
	bool is_null(seg_index seg) const;

	/**  Return the index of the last segment that starts at or before 
	  *  the specified gps.
	  *  @brief Find position in list for segment with gps t.
	  *  @param ichn Index of channel list to be searched.
	  *  @param gps  GPS time to be searched for.
	  *  @return Index of last segment with data before or at gps.
	  */
	seg_index last_leq(chan_index ichn, gps_type gps) const;

	/**  Move the specified segment to the tail (newest) position of the
	  *  age queue.
	  *  \note make_new is safe for inserting a segment that is not 
	  *        currently linked in the age queue.
	  *  \note The Age queue must not be locked when make_new is called. 
	  *  @brief Move segment to end of age queue. 
	  *  @param inx Index of segment to be moved.
	  */
	void make_new(seg_index inx);

	/**  Get a writable reference to the specified Segment.
	  *  @brief Get a segment reference.
	  *  @param seg index of the segment to be referenced.
	  *  @return Writeable reference to the specified segment.
	  */
	segment& operator[](seg_index seg);

	/**  Get a constant reference to the specified Segment.
	  *  @brief Get a segment reference.
	  *  @param seg index of the segment to be referenced.
	  *  @return Constant reference to the specified segment.
	  */
	const segment& operator[](seg_index seg) const;

	/**  Release a specified segment... Perform post processing e.g. 
	  *  merge with an adjacent segment if necessary/desired.
	  *  @brief Release a segment.
	  *  @param seg Index of segment to be released.
	  */
	void release(seg_index seg);

	/**  Reserve the segment containing the specified time for the 
	  *  specified channel and return the segment index. If a matching
	  *  segment is found, its use count is incremented and it is moved
	  *  to the end of the age list. If no segments are found with the
	  *  specified channel and time, the null index is returned.
	  *  @brief Reserve a segment.
	  *  @param chan Channel index
	  *  @param start Start time.
	  *  @return Index of first segment in the specified range.
	  */
	seg_index reserve(chan_index chan, gps_type start);

	/**  Reserve channel segment list headers for the specified number of 
	  *  channels. All channel entries must be inactive for this to work.  
	  *  In practice this means that the channel list size should be set 
	  *  before opening the server for client requests.  
	  *  \brief Set the number of channels.
	  *  \param nChans Number of channel entries to be allocated.
	  */
	void set_channel_count(chan_index nChans);

    private:
	/**  Find the first segment for the specified channel havning a start 
	  *  time greater than the specified gps.
	  *  @brief find a segment.
	  *  @param chan Channel index
	  *  @param gps Compare time.
	  *  @return Index of first segment after the specified time.
	  */
	seg_index first_gt_int(chan_index chan, gps_type gps) const;

	/**  Return the index of the last segment that starts at or before 
	  *  the specified gps. Internal entry point: does not read-lock 
	  *  the channel list.
	  *  @brief Find position in list for segment with gps t.
	  *  @param ichn Index of channel list to be searched.
	  *  @param gps  GPS time to be searched for.
	  *  @return Index of last segment with data before or at gps.
	  */
	seg_index last_leq_int(chan_index ichn, gps_type gps) const;

	/**  Insert the specified segment number into the age and channel 
	  *  lists. 
	  *  \note This method does not wait on the write mutex for either
	  *        the channel segment list or the segment age list. These
	  *        mutexes must be reserved before calling this function.
	  *  @brief Insert a segment. 
	  *  @param inx Index of segment to be inserted.
	  */
	void insert_int(chan_index chan, seg_index inx);

    private:
	seg_index mNSegs;

	//------------------------------  Age list and mutex
	qbase     mAgeQ;
	linkage   mAgeLinks;

	//-------------------------------  Channel list.
	base_vect mChanBase;
	linkage   mChanLinks;

	seg_vect  mSegVect;
    };

    //================================== Inline methods
    inline
    seg_db::linkage::linkage(seg_index N) 
	: mEnd(N), mVect(N, dll_node(N,N))
    {}
 
    inline void
    seg_db::linkage::unlink(seg_index inx) {
	seg_index ib4 = prev(inx);
	seg_index nxt = next(inx);
	if (ib4 != mEnd) mVect[ib4].second = nxt;
	if (nxt != mEnd) mVect[nxt].first  = ib4;
	mVect[inx].first  = mEnd;
	mVect[inx].second = mEnd;
    }

    inline void
    seg_db::linkage::insert_before(seg_index seg, seg_index iaft) {
	seg_index ib4 = prev(iaft);
	mVect[seg].first  = ib4;
	if (ib4 != mEnd) mVect[ib4].second = seg;
	mVect[iaft].first = seg;
	mVect[seg].second = iaft;
    }

    inline void
    seg_db::linkage::insert_after(seg_index seg, seg_index ib4) {
	seg_index iaft = mVect[ib4].second;
	mVect[seg].first  = ib4;
	mVect[ib4].second = seg;
	if (iaft != mEnd) mVect[iaft].first = seg;
	mVect[seg].second = iaft;
    }

    inline seg_index 
    seg_db::linkage::next(seg_index inx) const {
	return mVect[inx].second;
    }

    inline bool
    seg_db::linkage::null(seg_index N) const {
	return N == mEnd;
    }

    inline seg_index
    seg_db::linkage::prev(seg_index inx) const {
	return mVect[inx].first;
    }

    inline seg_index 
    seg_db::linkage::end(void) const {
	return mEnd;
    }

    inline
    seg_db::qbase::qbase(seg_index N) 
	: mFirst(N), mLast(N)
    {}

    inline
    seg_db::qbase::~qbase(void) { 
    }

    inline seg_index 
    seg_db::qbase::first(void) const {
	return mFirst;
    }

    inline seg_index 
    seg_db::qbase::last(void) const {
	return mLast;
    }

    inline seg_index 
    seg_db::qbase::next(seg_index iseg, const linkage& l) const {
	if (iseg == l.end()) return iseg;
	return l.next(iseg);
    }

    inline seg_index 
    seg_db::qbase::prev(seg_index iseg, const linkage& l) const {
	if (iseg == l.end()) return mLast;
	return l.prev(iseg);
    }

    inline thread::readwritelock& 
    seg_db::qbase::ref_mutex(void) const {
	return mRWMutex;
    }

    inline bool
    seg_db::chan_valid(chan_index chan) const {
	return chan < mChanBase.size();
    }

    inline seg_index
    seg_db::end(void) const {
	return mNSegs;
    }

    inline bool
    seg_db::is_null(seg_index inx) const {
	return inx == mNSegs;
    }

    inline segment& 
    seg_db::operator[](seg_index seg) {
	return mSegVect[seg];
    }

    inline const segment& 
    seg_db::operator[](seg_index seg) const {
	return mSegVect[seg];
    }

    //==================================  Define extern segment list.
    extern seg_db       Segment_List;

} // namespace sends

#endif  // !defined(SENDS_SEGDB_HH)
