/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_SEGMENT_HH
#define SENDS_SEGMENT_HH

#include "types.hh"
#include "gmutex.hh"
#include "data_type.hh"
#include <string>

namespace sends {

    class buffer;
    class pool_mgr;

    /**  The %segments class contains all the information about an existing 
      *  (or soon to exist) %segment of channel data.
      *  @brief Data %segment descriptor.
      *  @author J.Zweizig
      *  @version 1.0; Last modified December 31, 2007  
      *  @ingroup sends_base
      */
    class segment {
    public:
	/** Data type used for use count.
	  */
	typedef unsigned long count_type;

    public:
	/**  Construct a %segment descriptor.
	  *  @brief Constructor.
	  */
	segment(void);

	/**  Destroy a %segment descriptor.
	  *  @brief Segment destructor.
	  */
	~segment(void);

	/**  Get the bias value from the calibration transformation.
	  *  \brief Data bias.
	  *  \return Data bias.
	  */
	double bias(void) const;

	/**  Test whether the %segment can be taken for a buffer 
	  *  class specified. The %segment must be unused (mUseCount == 0) 
	  *  and have the sppropriate buffer class. A search buffer class 
	  *  of -1 selects a %segment with no allocated buffer.
	  *  @brief Test usable.
	  *  @param bufclass Class of buffer being searched for.
	  *  @return True if %segment can be used.
	  */
	bool canUse(int bufclass) const;

	/**  Set the maximum time for a %segment to a value less than the 
	  *  %segment capacity. This is used in cases where there is a gap 
	  *  in the data so that the data in the time between the current
	  *  end-time and the specified time can not be filled. The new 
	  *  maximum time must be greater than the reserve time (and hence 
	  *  the end time) and less than the current maximum time. If the
	  *  time is invalid by these criteria, a logic error is thrown.
	  *  @brief Cap a %segment at the specified time.
	  *  @param tMax New maximum time for the %segment.
	  */
	void cap(gps_type tMax);

	/**  Clear all fields of the %segment structure. Retain the buffer
	  *  (note that the buffer may be released with release).
	  *  @brief Clear a %segment.
	  */
	void clear(void);

	/**  Get the channel index for the channel being stored by this
	  *  %segment descriptor.
	  *  @brief Get channel index.
	  *  @return Index of channel whose data are contained in %segment.
	  */
	chan_index chan(void) const;

	/**  Get the index of the specified time in the data buffer.
	  *  @brief Calculate data position.
	  *  @param gps Time to be calculated.
	  *  @return Index to specified data.
	  */
	index_type dataIndex(gps_type gps) const;

	/**  Dump the status of the %segment.
	  *  @brief Dump %segment status.
	  *  @param out Output stream reference.
	  *  @return Output stream reference.
	  */
	std::ostream& dump(std::ostream& out) const;

	/**  Dump the %segment status header.
	  *  @brief Dump status header.
	  *  @param out Output stream reference.
	  *  @return Output stream reference.
	  */
	static std::ostream& dump_header(std::ostream& out);

	/**  Test whether the %segment buffer is undefined or empty.
	  *  @brief Test empty.
	  *  @return True if the %segment buffer is empty.
	  */
	bool empty(void) const;

	/**  The specified data are appended to those already in the data 
	  *  buffer starting at the fill pointer (mStop) and not exceding the
	  *  reserved pointer. The start of data offset (into the input buffer) 
	  *  and length are in words - the word length is specified at 
	  *  %segment initialization (set_segment method). The data bytes are
	  *  reordered as necessary so that the data are in network (big-
	  *  endian) byte order and read for transfer over the network.
	  *  No mutexes are cleared.
	  *  \brief Extend data.
	  *  \param data   Data to be appended to the %segment.
	  *  \param inx    Word index into 'data' of data to be added to the 
	  *                %segment.
	  *  \param length Number of data words to be added to the %segment.
	  *  \return True if reserved space has been filled.
	  */
	bool extend(const void* data, count_type inx, count_type length);

	/**  Unreserve space in the %segment and release the fill mutex.
	  *  \exception logic_error if the fill mutex is not locked.
	  *  \brief Fill failure processing.
	  */
	void fill_failed(void);

	/**  Increment the end pointer by the specified time (\a dt). If the
	  *  new stop time is equal to the reserve pointer, unlock the fill 
	  *  mutex and return true. 
	  *  \exception logic_error if the fill mutex is not locked.
	  *  \brief Increment the fill pointer.
	  *  \param dt number of seconds to increment the fill pointer.
	  *  \return true if the new fill pointer (stop time) is equal to 
	  *               the last reserved data time.
	  */
	bool fill_increment(int dt);

	/**  Lock the %segment to add data. The data position and currently
	  *  existing data (as defined by the start and stop times) remain 
	  *  available for read access.  No manipulation of the stop time,
	  *  or the reserved data time should occur without first issuing
	  *  a fill_lock request.
	  *  @brief Lock the %segment for filling.
	  */
	void fill_lock(void);

	/**  Test whether the %segment was previously locked for filling and
	  *  throw an exception if the lock is off.
	  *  \exception logic_error if fill is not locked.
	  *  \brief Test fill lock;
	  *  \param name Name of calling function.
	  */
	void fill_test(const std::string& name);

	/**  Wait for fill lock to complete (lock and unlock the fill.
	  *  @brief Wait for fill to complete.
	  */
	void fill_wait(void);

	/**  Test whether the %segment buffer is full.
	  *  @brief Test full.
	  *  @return True if full.
	  */
	bool full(void) const;

	/**  Return the buffer class code of the buffer allocated for this 
	  *  %segment. If no buffer is allocated, the class code is set to
	  *  -1.
	  *  @brief Get buffer class.
	  *  @return Buffer class code.
	  */
	int getBufClass(void) const;

	/**  Get the GPS time of the first sample not currently available 
	  *  in the buffer.
	  *  @brief Get end time.
	  *  @return Data end time.
	  */
	gps_type getEndGPS(void) const;

	/**  Get buffer full time.
	  *  @brief Get max time.
	  *  @return Buffer max time.
	  */
	gps_type getMaxGPS(void) const;

	/**  Get data reserve time.
	  *  @brief Get sreserve time.
	  *  @return Data reserve time.
	  */
	gps_type getReserveGPS(void) const;

	/**  Get GPS time of the first sample in the %segment buffer.
	  *  @brief Get start time.
	  *  @return Data start time.
	  */
	gps_type getStartGPS(void) const;

	/**  Get the current use count. The getUseCount method is not 
	  *  implicitly sychronized using a mutex so unless the caller
	  *  issues a read-lock the count may be modified at any time after
	  *  it is used.
	  *  @brief Get use count.
	  *  @return Use count.
	  */
	count_type getUseCount(void) const;

	/**  Test whether the specified time is within the range covered 
	  *  by this %segment, \e i.e. start \c \<= \a gps \c \< stop.
	  *  @brief Test if time is in this %segment.
	  *  @param gps Time to be tested.
	  *  @return True if specified time is in the %segment time range.
	  */ 
	bool in_segment(gps_type gps) const;

	/**  Test whether this %segment is in use, i.e. if the use count is 
	  *  non-zero.
	  *  @brief Test if the %segment is in use.
	  *  @return True if the %segment is in use.
	  */
	bool inUse(void) const;

	/**  Compare the times of two %segment objects. The l.h.s. %segment is
	  *  less if its stop time is before or equal to the start time of the 
	  *  r.h.s. %segment.
	  *  @brief compare time order of two segments.
	  *  @param seg Segment to be compared to this.
	  *  @return True if this %segment is before the argument %segment.
	  */
	bool operator<(const segment& seg) const;

	/**  Return the sample rate in Hz.
	  *  @brief Sample rate
	  *  @return Sample rate in Hz.
	  */
	double rate(void) const;

	/**  Return a constant pointer to the data item at the specified
	  *  GPS time.
	  *  @brief Return pointer to specified data.
	  *  @param gps Gps time of requested data.
	  *  @return Constant pointer to data sample at specified time.
	  */
	const char* refData(gps_type gps) const;

	/**  Return a pointer to the data item at the specified
	  *  GPS time.
	  *  @brief Return pointer to specified data.
	  *  @param gps Gps time of requested data.
	  *  @return Pointer to data sample at specified time.
	  */
	char* refData(gps_type gps);

	/**  Release the buffer and clear the %segment.
	  *  @brief Release and clear the %segment.
	  */
	void release(void);

	/**  Extend buffer space to the specified time by incrementing the
	  *  reserved pointer. If the start time is after the current %segment
	  *  fill point, the reserved pointer is not incremented. If the 
	  *  %segment will not accomodate the entire interval, the reserved
	  *  pointer is set to the maximum time.
	  *  @brief Extend reserved buffer space.
	  *  @param start start of reserved interval.
	  *  @param stop Time at end of buffer space.
	  *  @return true if the reserved pointer was incremented.
	  */
	bool reserve(gps_type start, gps_type stop);

	/**  Revert the reserve time to the end of available data and release 
	  *  the %segment for read access.
	  *  \brief Revert the data reservation
	  */
	void revert(void);

	/**  Get the scale factor from the calibration transformation.
	  *  \brief Data scale.
	  *  \return Data scale.
	  */
	double scale(void) const;

	/**  Set the buffer pointer and buffer type to the argument values.
	  *  @brief Set the buffer.
	  *  @param bufpt  Buffer address.
	  *  @param btype Buffer type code.
	  */
	void set_buffer(buffer* bufpt, int btype);

	/**  Set the buffer content information to as appropriate for the
	  *  specified channel and start-time.
	  *  @brief Set up a buffer
	  *  @param chan  Index of channel to be assigned to the %segment.
	  *  @param rate  Channel sample rate
	  *  @param dtyp  Channel data word type
	  *  @param start Start time of %segment
	  */
	void set_segment(chan_index chan, double rate, 
			 data_type dtyp, gps_type start);

	/**  Set the linear data transformation and resulting units.
	  *  \brief Specify units.
	  *  \param bias  Data bias.
	  *  \param scale Data scale factor
	  *  \param units Calibrated units.
	  */
	void set_units(double bias, double slope, const std::string& units);

	/**  Get the number of seconds of data that can be accomodated 
	  *  at the end of the buffer.
	  *  \brief Remaining data length.
	  *  \return Remaining length in seconds.
	  */
	gps_type time_left(void) const;

	/**  Increment the use count for this %segment, and return the previous
	  *  use count (\e i.e. the count before this call).
	  *  \brief Use the %segment.
	  *  \return Previous use count.
	  */
	count_type use(void);

	/**  Get the units name string
	  *  \brief Unit name string
	  *  \return Units name string reference;
	  */
	const std::string& units(void) const;

	/**  Decrement the use count for this %segment in a thread-safe manner. 
	  *  Return the remaining use count.
	  *  @brief Unuse the %segment.
	  *  @return Number of transactions still requiring this %segment.
	  */
	count_type unuse(void);

    private:
	thread::mutex mFillMutex;
#ifndef USE_BUILTIN_ATOMICS
	thread::mutex mUseMutex;
#endif
	int        mBufClass;
	chan_index mChannel;
	gps_type   mStart;
	gps_type   mStop;
	gps_type   mReserved;
	gps_type   mTMax;
	double     mRate;
	data_type  mWordType;
	count_type mUseCount;
	buffer*    mData;
	double     mBias;
	double     mScale;
	std::string mUnits;
    };

    //================================== Inline methods
    inline double
    segment::bias(void) const {
	return mBias;
    }

    inline bool 
    segment::canUse(int bufclass) const {
	return !mUseCount && mBufClass == bufclass;
    }

    inline chan_index
    segment::chan(void) const {
	return mChannel;
    }

    inline bool
    segment::empty(void) const {
	return mStop == mStart;
    }

    inline bool
    segment::full(void) const {
	return mStop == mTMax;
    }

    inline int
    segment::getBufClass(void) const {
	return mBufClass;
    }

    inline gps_type 
    segment::getEndGPS(void) const {
	return mStop;
    }

    inline gps_type 
    segment::getMaxGPS(void) const {
	return mTMax;
    }

    inline gps_type 
    segment::getReserveGPS(void) const {
	return mReserved;
    }

    inline gps_type 
    segment::getStartGPS(void) const {
	return mStart;
    }

    inline segment::count_type 
    segment::getUseCount(void) const {
	return mUseCount;
    }

    inline bool 
    segment::in_segment(gps_type gps) const {
	return gps>=mStart && gps<mStop;
    }

    inline bool 
    segment::inUse(void) const {
	return mUseCount;
    }

    inline bool
    segment::operator<(const segment& seg) const {
	return mStop <= seg.mStart;
    }

    inline double 
    segment::rate(void) const {
	return mRate;
    }

    inline double
    segment::scale(void) const {
	return mScale;
    }

    inline gps_type 
    segment::time_left(void) const {
	return mTMax - mStop;
    }

    inline const std::string&
    segment::units(void) const {
	return mUnits;
    }

} // namespace sends

#endif  // !defined(SENDS_SEGMENT_HH)
