/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_MAIN_HH
#define SENDS_MAIN_HH

#include "authorize.hh"
#include "concentrator.hh"
#include "parse_stage.hh"
#include "lookup_stage.hh"
#include "fetch_stage.hh"
#include "write_stage.hh"

namespace sends {

    class transaction;

    /**  Process class of the SENDS server.
      *  @brief SENDS server class.
      *  @author John Zweizig
      *  @version 1.0; Last Modified January 23, 2008
      *  @ingroup sends_main
      */
    class sends_main {
    public:
	/**  Construct the server class. The constructor includes 
	  *  initialization and interconnection of the sends 
	  *  pipeline stages and their thread pools, and the initialization
	  *  of the use, channel, frame and segment databases.
	  *  @brief Constructor.
	  *  @param argc Number of command line arguments
	  *  @param argv Vector of command line argument pointers.
	  */
	sends_main(int argc, const char* argv[]);

	/**  Shut down server.
	  *  @brief Destructor.
	  */
	~sends_main(void);

	/**  Read configuration file.
	  *  \brief configure.
	  *  \param file Configuration file name.
	  */
	void configure(const std::string& file);

	/**  Read configuration file.
	  *  \brief configure.
	  *  \param file Configuration file name.
	  */
	void check_config(void) const;

	/**  Server processing method.
	  *  @brief Main routine.
	  */
	void doit(void);

	/**  Test initialization complete
	  *  \brief Test initialization
	  *  \return True if initialization complete.
	  */
	bool init_complete(void) const;

	/**   Print all transactions.
	 */
	void dump_transactions(std::ostream& out) const;

	/**  Print transactions from one stage
	 */
	void dump_pool(int id, const seda_stage& stage, std::ostream& out) const;

    private:
	//------------------------------  Parameter storage data types
	typedef std::map<std::string, seda_stage::stage_pars> stage_par_map;
	typedef stage_par_map::iterator       stage_par_iter;
	typedef stage_par_map::const_iterator const_stage_par_iter;
	typedef stage_par_map::value_type     stage_par_node;

    private:
	//------------------------------  Configuration file
	std::string  mConfigure;
	bool         mInitComplete;

	//------------------------------  Stages
	authorize    mAuthPool;
	concentrator mConcPool;
	parse_stage  mParsePool;
	lookup_stage mLookupPool;
	fetch_stage  mFetchPool;
	write_stage  mWritePool;
 	tran_queue   tranQ;
	std::vector<transaction*> tranVec;

	//------------------------------  Parameters
	stage_par_map mStageParMap;
	int           mSmallBufSize;
	int           mSmallBufCount;
	int           mLargeBufSize;
	int           mLargeBufCount;
	int           mXLargeBufSize;
	int           mXLargeBufCount;
	int           mTransCount;
	std::string   mListenAddr;
	int           mListenQueue;
	typedef std::pair<unsigned int, unsigned int> allow_node;
	std::vector<allow_node> mAllowVect;
	auth_server::auth_type  mAuthType;
    };

    //==================================  Inline methods
    inline bool
    sends_main::init_complete(void) const {
	return mInitComplete;
    }
}

#endif  // !defined(SENDS_MAIN_HH)
