/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SORTED_VECTOR_HH
#define SORTED_VECTOR_HH
#include <vector>
#include <functional>

/**  Vector based template with sort, and ordered insertion and merging
  *  extensions. 
  *  @memo Sorted vector template
  *  @author John Zweizig
  *  @version 1.0; Last modified 2009-10-13
  *  @ingroup sends_base
  */
template<typename _Tp, typename _Cmp = std::less<_Tp>, 
	 typename _Alloc = std::allocator<_Tp> >
class sorted_vector : public std::vector<_Tp, _Alloc>
{
public:
    //----------------------------------  Define internal types
    /**  Base vector type
      */
    typedef          std::vector<_Tp, _Alloc>  base_type;

    /**  Iterator of base vector type
      */
    typedef typename base_type::iterator       iterator;

    /**  Constant iterator of base vector type
      */
    typedef typename base_type::const_iterator const_iterator;

    /**  Size type
      */
    typedef typename base_type::size_type      size_type;

    /**  Value  type
      */
    typedef typename base_type::value_type     value_type;

public:
    /**  Find (the first occurrance of) the specified value in the vector.
      *  A binary search is used to optimize the search. If the value is not 
      *  found, the iterator is set to end().
      *  @memo find an entry containing the specified value.
      *  @param val Value to be found.
      *  @return Iterator pointing to an entry containing the value or end().
      */
    iterator find(const value_type& val);

    /**  Find (the first occurrance of) the specified value in the vector.
      *  A binary search is used to optimize the search. If the value is not 
      *  found, the iterator is set to end().
      *  @memo find an entry containing the specifie value.
      *  @param val Value to be found.
      *  @return Constant iterator pointing to the first entry containing 
      *          the value or end().
      */
    const_iterator find(const value_type& val) const;

    /**  Find (the first occurrance of) the specified value in the vector.
      *  A binary search is used to optimize the search. If the value is not 
      *  found, the iterator is set to the size().
      *  @memo find an entry containing the specifie value.
      *  @param val Value to be found.
      *  @return Index of the first entry containing the value or end().
      */
    size_type findex(const value_type& val) const;



    /**  Return a constant iterator pointing to the first entry that is not
      *  less than the argument value.
      *  @memo Find the lower bound
      *  @param val Comparison value.
      *  @return Constant iterator pointing to the appropriate vector position.
      */
    const_iterator lower_bound(const value_type& val) const;

    /**  Return an iterator pointing to the first entry that is not less 
      *  than the argument value.
      *  @memo Find the lower bound
      *  @param val Comparison value.
      *  @return Constant iterator pointing to the appropriate vector position.
      */
    iterator lower_bound(const value_type& val);

    /**  Merge an unsorted vector of the same type into the sorted
      *  vector.
      *  @memo Merge vectors.
      *  @param vec Unsorted vector of the same type.
      */
    void merge(const base_type& vec);

    /**  Return a constant iterator pointing to the first entry that is 
      *  greater than the specified value.
      *  @memo Find the upper bound
      *  @param val Comparison value.
      *  @return Constant iterator pointing to the appropriate vector position.
      */
    const_iterator upper_bound(const value_type& val) const;

    /**  Find the first entry in the vector that is greater than the 
      *  argument value.
      *  @memo Find the upper bound
      *  @param val Comparison value.
      *  @return Iterator pointing to the appropriate vector position.
      */
    iterator upper_bound(const value_type& val);

    /**  Insert the specified data item at the appropriate position of
      *  the vector.
      *  @memo Insert a new entry.
      *  @param val entry to be inserted into the vector.
      */
    void insert(const value_type& val);

    /**  Reorder the vector contents in the order defined by the _Cmp 
      *  template argument.
      *  \brief Sort the vector.
      */
    void sort(void);
};

//======================================  Include the inline templates
// #define INT_SORT
#include "sortedvector.icc"

#endif // !defined(SORTED_VECTOR_HH)
