/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_SOURCE_HISTO_HH
#define SENDS_SOURCE_HISTO_HH

#include "types.hh"
#include "sortedvector.hh"
#include <string>
#include <vector>

namespace sends {

    class loc_list;
    class chan_entry;

    /**  The source information structure contains any information sbout 
      *  a single source type needed by the histogrammer.
      *  \brief Source information structure.
      *  \author  John Zweizig
      *  \version $Id: source_histo.hh 642 2012-03-28 20:43:53Z john.zweizig@LIGO.ORG $
      */
    struct source_info {
	/** Initialize the source information.
	  * \brief Source info constructor.
	  * \param name Name of the source.
	  */
	source_info(const std::string& name) 
	    : _name(name), _nchn(0), _arch(0), _loc(0) 
	{}

	/**  Compare use counts and archive type of two source_info 
	  *  structures. Return true if this is used less than the argument
	  *  structure.
	  *  \param x source_info structure to which this is compared.
	  *  \return True if this is less than the argument.
	  */
	bool operator<(const source_info& x) const;

	/**  Release the frame_loc list - we don't need it any more.
	  *  \brief Release frame_loc.
	  */
	void release(void);

	/**  Swap source info data with those in the argument instance.
	  *  \brief Swap source info data.
	  *  \param x Instance woth which to swap data.
	  */
	void swap(source_info& x);

	/** Source name
	  */
	std::string _name;

	/**  Number of channels available from this source.
	 */
	int         _nchn;

	/**  Source archival type.
	 */
	int         _arch;

	/**  Frame location list for this source.
	 */
	const loc_list* _loc;
    };

    //==================================  Inline methods.
    inline bool 
    source_info::operator<(const source_info& x) const {
	return _nchn < x._nchn || (_nchn == x._nchn && _arch > x._arch);
    }

    /**  The source_histo class is used to select the most promising 
      *  data source, i.e the source that is most rapidly accessible 
      *  and provides the maximum number of requested channels.
      *  \brief Histogram and sort a list of sources.
      */
    class source_histo {
    public:
	/**  Construct a source histogram.
	 */
	source_histo(void);

	/**  Destroy a source histogram.
	  *  \brief Destructor
	  */
	~source_histo(void);

	/**  Increment the use count for all sources from which the specified
	  *  channel is available.
	  *  \brief Bump use counts for a channel.
	  *  \param chan Channel list entry for the channel to be processed.
	  *  \param t    Time at which use is to be evaluated.
	  */
	void bump(const chan_entry& chan, gps_type t);

	/**  Increment the use count for the specified data source.
	  *  \brief Bump use counts for a single source.
	  *  \param name Name of source to be processed.
	  *  \param t    Time at which use is to be evaluated.
	  */
	void bump(const std::string& name, gps_type t);

	/**  Remove all sources from the histogram.
	  *  \brief Clear the histogram.
	  */
	void clear(void);

	/**  Test if the histogram is empty.
	  *  \return true if no valid sources.
	  */
        bool empty(void) const;

	/**  Find the histogram entry for the specified source.
	  *  \param name Nam of source to be found
	  *  \return Index of the named source.
	  */
	index_type find(const std::string& name) const;

	/**  Get a reference to the first (most promising) source.
	  *  \brief Reference best source info.
	  *  \return Index of the named source.
	  */
	source_info& first(void);

	/**  Test whether the specified source has been vetoed.
	  *  \brief Test source veto.
	  *  \param src Source name to be tested.
	  *  \return True if the source has been vetoed.
	  */
	bool is_vetoed(const std::string& src) const;

	/**  Number of sources in the histogram.
	  *  \return number of sources.
	  */
	index_type size(void) const;

	/**  Add a source name to the veto list. The veto list is checked 
	  *  before incrementing the use count 
	  *  \brief Veto a source
	  *  @param src Source to be vetoed.
	  */
	void veto(const std::string& src);

    private:
	/** Vector of source_info structures.
	  */
	typedef std::vector<source_info> arch_vect;

	/**  Source use histogram.
	  */
	arch_vect av;

	/**  Veto list.
	  */
	sorted_vector<std::string> uv;
    };

    //==================================  Inline methods
    inline bool 
    source_histo::empty(void) const {
	return av.empty();
    }

    inline source_info&
    source_histo::first(void) {
	return av[0];
    }

    inline bool
    source_histo::is_vetoed(const std::string& src) const {
	return !uv.empty() && uv.find(src) != uv.end();
    }

    inline index_type 
    source_histo::size(void) const {
	return av.size();
    }

}  //  namespace sends

#endif  // !defined(SENDS_SOURCE_HISTO_HH)
