/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_SRC_UPDATE_HH
#define SENDS_SRC_UPDATE_HH

#include "types.hh"
#include "frame_source.hh"
#include "thread_base.hh"
#include "gateway.hh"

namespace sends {

    /**  The source update class updates the source data availability. It 
      *  can run by testing specific frame names or by testing a time-stamp 
      *  file with the start of the latest frame.
      *  The update class is used to maintain current data in the nds2 source
      *  lists. Each updater is defined by a source.update line in the nds2
      *  configuration file and may run in one of the following modes:
      *  <ul>
      *  <li>scan: The frame data list is regenerated by scanning through 
      *            the directory path descriptor</li>
      *  <li>sched: A the presence of data frames is determined by a binary
      *             search from the current end-time to the most recent 
      *             expected frame time.</li>
      *  <li>synch: The end time is updated from the spcified time-stamp 
      *             file.</li>
      *  </ul>
      *  \brief Update source lists.
      *  \author John Zweizig
      *  \version 1.0; Last modified February 16, 2009
      *  \ingroup sends_base
      */
    class src_update : public thread::thread_base {
    public:

	/** Enumerated update scheduling modes.
	  */
	enum sched_mode {
	    kNone,    ///< Scheduling mode not set.
	    kSched,   ///< Scheduled updates.
	    kSynch,   ///< File-synchronized scheduling.
	    kScan     ///< Rescan directory(ies)
	};

	/** Message state code used to suppress repeated messages.
	 */
	enum msg_state {
	    /// normal running state
	    kNormal,
	    /// no synch file message printed
	    kNoSync
	};

    public:
	/**  Construct an updater.
	  */
	src_update(void);

	/**  Destroy an updater.
	  */
	~src_update();

	/**  Use a binary search to find the current start of available 
	  *  data for a specified data source.
	  *  \param ll Frame locator for data source to search.
	  *  \param lo Lower GPS bound for search.
	  *  \param hi Upper GPS bound for search.
	  *  \return Start of available data.
	  */
	gps_type beg_time(const loc_list& ll, gps_type lo, gps_type hi) const;

	/**  Return the start time of the most recent file.
	  *  \brief current time
	  *  \return The most recent file time.
	  */
	gps_type current_time(void) const;

	/**  Use a binary search to find the current end of data for the 
	  *  specified frame source. Note that the returned value is the
	  *  file start time, indicating that the data extends past the
	  *  return value by the duration of the frame. The search range, 
	  *  specified by lo - hi, should start with the last known data.
	  *  \brief Find the current end of data.
	  *  \param fl Frame loc for which the data are to be searched.
	  *  \param lo Low limit of the time s to be searched for available
	  *  \param hi Upper limit on times to be searched.
	  *  \return Start time of the last found available file.
	  */
	gps_type end_time(const loc_list& ll, gps_type lo, gps_type hi) const;

	/**  Return descriptive name in the format "<arch-type>:<frame-type>".
	 */
	std::string name(void) const;

	/**  Use a binary search to find the current end of data.
	  */
	void now_available(void);

	/**  Set the debug print verbosity flag
	  *  \brief Set the debug flag
	  *  \param dbg Debug print level.
	  */
	void  set_debug(int dbg);

	/**  Set the scan mode and time parameters. The scan periodically 
	  *  rescans the frame directories. The rescan occurs with a period 
	  *  specified by the latency argument. 
	  *  \param latency Schedule delay after next expected frame gps time.
	  *  \param retry   Schedule delay after failed test.
	  */
	void  set_scan_mode(double latency=0, double retry=0);

	/**  Set the scheduled mode and time parameters. The scheduled mode
	  *  makes periodic updates the available data range. By default,
	  *  the updates occur with a period of the frame file length, 
	  *  aligned to the integer multiples of that time.
	  *  \param latency Schedule delay after next expected frame gps time.
	  *  \param retry   Schedule delay after failed test.
	  */
	void  set_sched_mode(double latency=0, double retry=0);

	/**  Set the synchronized scheduling file, mode and time parameters. 
	  *  In synchronized mode the updater polls the specified file
	  *  for time of the most recently written frame.
	  *  \param file Synchronization file name.
	  *  \param late Schedule delay after next expected frame gps time.
	  *  \param step Schedule delay after failed test.
	  */
	void  set_synch_mode(const std::string& file, double late=0, 
			     double step=0.5);

	/**  Set the frame location to be updated.
	  *  \param a Archival type of source to be updated.
	  *  \param s Name of source to be updated.
	  *  \param t Gps time in source segment to be updated.
	  */
	void  set_source(loc_list::arch_type a, const std::string& s,
			gps_type t);

	/** Get the source type
	 */
	const std::string& source_type(void) const;
	
	/**  Pre-configure and start the thread.
	 */
	int   start(void);

	/**  Stop the thread.
	 */
	int   stop(void);

	/**  Updater thread polling code. This method runs asynchronously
	  *  as a thread and polls the specified data location for new data.
	  *  The polling method is specified by the sched_mode enumerator.
	  *  \brief Asynchronous source polling.
	  *  \exception std::runtime_error Unable to locate a frame location 
	  *             (frame_loc) for the specified data source.
	  *  \return Thread return value (zero for normal termination).
	  */
	void* thread_entry(void);

	/**  Wait allows an external to wait thread a maximum time interval 
	  *  for data from the updated source to become available. If \a tmax
	  *  is negative, the thread will wait indefinitely for an update.
	  *  \brief Wait for the next update.
	  *  \param tmax Maximum time (in seconds) to wait for an update.
	  *  \return true if an update has occurred.
	  */
	bool  timed_wait(double tmax) const;

	/**  Wait allows an external thread to wait for data from the updated
	  *  source to become available.
	  *  \brief Wait for the next update.
	  */
	void  wait(void) const;

	/**  Wait for the source to be updated beyond the specified time.
	  *  I \a tmax is  negative, the thread will wait indefinitely for
	  *  the requested data. Otherwise, the thread will wait a maximum 
	  *  of \a tmax seconds.
	  *  \param t    Earliest gps time needed.
	  *  \param tmax Maximum wait time.
	  *  \return True if source data are available for the specified time.
	  */
	bool  wait(gps_type t, double tmax=-1) const;

    private:
	/**  Test whether a frames exists for this source at the specified 
	  *  time.
	  *  \brief Test if frame is available.
	  *  \param fl Frame location specifier
	  *  \param t New time.
	  */
	bool test_file(const frame_loc& fl, gps_type t) const;

	/**  Update the current time in the frame_loc element and release 
	  *  any threads waiting for an update;
	  *  \brief perform update processing
	  *  \param ll Frame location specifier list
	  *  \param t New time.
	  */
	void update(loc_list& ll, gps_type t);

    private:
	//----------------------------------  Target source parameters
	loc_list::arch_type  mArchive;
	std::string          mSrcType;
	volatile gps_type    mCurrentTime;

	//----------------------------------  Scheduling parameters
	sched_mode mSchedMode;
	double     mLatency;   // relative time to start polling in synch mode
	double     mStep;      // polling interval.
	std::string mSynchFile;
	int        mDebug;     // debug flag

	//----------------------------------  Process statistics
	msg_state  mMsgState;
	double     mSumLate;   // sum of measured latencies
	long       mNMeasure;  // Number of latency measurements.

	thread::readwritelock   mLock;
	mutable thread::gateway mWaitGateway;
	mutable thread::gateway mSynchEvent;
    };

    //==================================  Inline methods.
    inline gps_type
    src_update::current_time(void) const {
	return mCurrentTime;
    }
    inline const std::string&
    src_update::source_type(void) const {
	return mSrcType;
    }
}

#endif // !defined(SENDS_SRC_UPDATE_HH)
