/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_THREAD_POOL_HH
#define SENDS_THREAD_POOL_HH
#include <pthread.h>
#include "gateway.hh"


namespace sends {

    /**  The thread pool base class contains the API of the thread 
      *  procedure to be performed by each thread in the pool and status 
      *  and control information for the pool as a whole, including:
      *  <ul>
      *    <li>A mechanism to change number of running threads</li>
      *    <li>A busy counter: number of busy threads</li>
      *    <li>A use counter: total number of events processed by the pool</li>
      *    <li>Mutex lock to protect status information</li>
      *  </ul>
      *  @brief Thread pool base class.
      *  @author John Zweizig
      *  @version 1.0; Last modified January 23, 2008
      *  @ingroup sends_base
      */
    class thread_pool {
    public:
	/**  This method is used to start up the 
	  *  thread class.
	  *  @brief Thread start method.
	  *  @param tp Pointer to the class to be 
	  *  @return Thread return value.
	  */
	static void* thread_exec(void* tp);

	/**  Data type used to specify thread counts.
	  *  @brief Thread count data type.
	  */
	typedef unsigned int count_type;

    public:
	/**  Construct a thread pool, specifying the number of threads to be
	  *  run concurrently. Threads aren't created until the user calls
	  *  make threads after completing any initialization needed.
	  *  @brief Thread pool constructor.
	  *  @param num Number of threads to be run concurrently.
	  */
	thread_pool(count_type num);

	/**  Destroy a thread pool.
	  *  @brief Destructor.
	  */
	virtual ~thread_pool(void);

	/**  Get the current maximum number of threads to run concurrently.
	  *  @brief Get the maximum number of threads.
	  *  @return Number of threads.
	  */
	count_type capacity(void) const;

	/**  Return the number of events the threads have processed. 
	  *  This counter is incremented by each transition of a thread to
	  *  the busy state.
	  *  @return Number of processing events.
	  */
	count_type count(void) const;

	/**  Thread pool processing method.
	  *  @brief Process method.
	  */
	virtual void thread_entry(void);

	/**  Number of threads in the busy state.
	  *  @brief Number of busy threads.
	  *  @return Number of busy threads.
	  */
	count_type get_busy(void) const;

	/**  High water mark for threads in the busy state.
	  *  @brief Maximum number of busy threads.
	  *  @return Maximum number of busy threads.
	  */
	count_type get_max_busy(void) const;

	/**  Lock the thread pool status information for modification.
	  *  @brief Lock the thread pool.
	  */
	void lock(void);

	/**  Create as many threads as necessary to bring the current number 
	  *  to max_threads.
	  *  @brief Create threads.
	  */
	void make_threads(void);

	/**  Set the maximum number of threads.
	  *  @brief Set number of threads.
	  *  @param N Number of threads.
	  */
	void set_size(count_type N);

	/**  Get the current number of running threads.
	  */
	count_type size(void) const;

	/**  Unlock the thread pool status and control information.
	  *  @brief Unlock the thread pool.
	  */
	void unlock(void);

	/**  Wait for the thread pool to empty out.
	  *  @brief Wait for the end
	  */
	void wait_empty(void);

    protected:
	/**  Indicate the the current thread is busy. The busy count and
	  *  use count are incremented.
	  *  @brief Current thread is busy.
	  *  @return Value of the use count before starting this thread.
	  */
	count_type busy(void);

	/**  Indicate the the current thread has finished. The busy count
	  *  is decremented.
	  *  @brief Current thread is done.
	  */
	void done(void);

	/**  test whether the thread is to continue running. Note that the 
	  *  thread must test its run state periodically and terminate 
	  *  immediately if the return value is false.
	  *  \return true if thread to continue.
	  */
	bool test_run(void);

    private:
	thread::gateway gate;   ///< Mutex to protect the status info
	count_type max_threads; ///< Maximum number of threads
	count_type nthread;     ///< Number of threads currently in pool
	count_type nbusy;       ///< Number of busy threads in pool

	//-----------------------------  Statistics
	count_type use_count;   ///<  Number of threads dispatched from pool 
	count_type max_busy;    ///<  Maximum number of busy threads 
    };

    //=================================  Inline methods.
    inline  thread_pool::count_type
    thread_pool::get_busy(void) const {
	return nbusy;
    }

    inline  thread_pool::count_type
    thread_pool::get_max_busy(void) const {
	return max_busy;
    }

    inline  thread_pool::count_type
    thread_pool::count(void) const {
	return use_count;
    }

    inline  thread_pool::count_type
    thread_pool::capacity(void) const {
	return max_threads;
    }

    inline  thread_pool::count_type
    thread_pool::size(void) const {
	return nthread;
    }

}

#endif // !defined(SENDS_THREAD_POOL_HH)
