/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_TRANSACTION_HH
#define SENDS_TRANSACTION_HH

//#include "tbb/concurrent_queue.h"
#include "mt_queue.hh"
#include "types.hh"
#include "socket_api.hh"
#include "auth_server.hh"
#include "chan_list.hh"
#include <vector>
#include <iosfwd>

//=====================================  NDS2 response codes copied from client
#include "daqc_response.h"

/** The server protocol version is 1 for nds2.
  */
const int SERVER_PROTOCOL_VERSION = 1;

/**  The server protocol revisions are:
  *
  *  Rev  Date     by  Changes
  *   3  9/24/14  JGZ  Add channel name pattern matching.
  *   4  4/15/15  JGZ  Add check-data command.
  *   5  2/21/16  JGZ  Add get-last-message command.
  *   6  7/14/17  JGZ  Negotiated protocol, add units.
  */
const int SERVER_PROTOCOL_REVISION = 6;

namespace sends {

    class transaction;
    class buffer;
    class driver_api;
    class tran_queue;  //see below

    /**  SENDS transaction block containing the SSL connection infomation,
      *  the user information, status information and other junk. The 
      *  transaction structure is passed between the SEDA stages as the 
      *  transaction is processed by the pipeline.
      *  @brief SENDS Transaction structure
      *  @author J. Zweizig
      *  @version 1.3; Last Modified May 13, 2009
      *  @ingroup sends_stage
      */
    class transaction {
    public:

	/**  The chan_stat class contains the channel ID and current processing
	  *  status of each channel involved in the transaction.
	  *  @brief Channel ID and status.
	  *  @ingroup sends_stage
	  */
	class chan_stat {
	    chan_index  _chinx;
	    gps_type    _fillpt;
	    int         _status;
	    driver_api* _driver;
	public:
	    /**  Construct a channel status object for the specified channel.
	      *  @brief Channel constructor.
	      *  @param chn Index of the channel into the channel list.
	      */
	    chan_stat(chan_index chn);

	    /**  Null channel constructor. This constructor is made available
	      *  for use when constructing the transaction channel list.
	      *  @brief Default Constructor.
	      */
	    chan_stat(void);

	    /**  Test the channel availablility.
	      *  @brief Test channel available
	      *  @return true if data available.
	      */
	    bool available(void) const;

	    /**  Get a pointer to the driver from which the channel was 
	      *  requested.
	      *  @memo Get driver pointer
	      *  @return Pointer to driver api.
	      */
	    driver_api* driver(void) const;

	    /**  Get the fill point for the channel. The fill point is the 
	      *  first second of data after the transactoin start time not 
	      *  yet reserved for the transaction.
	      *  @brief Fill time GPS.
	      *  @return GPS time to which data have been reserved
	      */
	    gps_type fill(void) const;

	    /**  Return the index of the channel.
	      *  @brief Channel index
	      *  @return Index of the channel in the master channel list.
	      */
	    chan_index index(void) const;

	    /**  Test whether all available data for channel in the current 
	      *  stride have been read into segments and reserved. If the 
	      *  status indicates data are unavailable or protected ready 
	      *  will be true.
	      *  @brief Test whether all available channel data are reserved.
	      *  @param stop End of the current stride.
	      *  @return True if all available data in stride are reserved.
	      */
	    bool ready(gps_type stop) const;

	    /**  Set the driver address.
	      *  @brief Set the driver address.
	      *  @param d pointer to the driver on which channel is registered.
	      */
	    void set_driver(driver_api* d);

	    /**  Set the fill time to the specified GPS.
	      *  @brief Set the fill time.
	      *  @param t time to which the fill pointer is set.
	      */
	    void set_fill(gps_type t);

	    /**  Set the error status word. The error status word has the 
	      *  following values:
	      *  - \em 0: OK
	      *  - \em -1: Requested data not available.
	      *  - \em -2: Access prohibited.
	      *
	      *  @brief Set the error status.
	      *  @param e Error status
	      */
	    void set_stat(int e);

	    /**  Get the error status code. The error status word has the 
	      *  following values:
	      *  - \em 0: OK
	      *  - \em -1: Requested data not available.
	      *  - \em -2: Access prohibited.
	      *
	      *  @brief Get the fill error status.
	      *  @return Status code
	      */
	    int status(void) const;
	};

	/**  Data type for list (vector) oc channel status objects.
	  */
	typedef std::vector<chan_stat> cinx_vect;

	/**  Data type used to maintain a list of reserved segments.
	  *  @brief Reserved segment list type.
	  */ 
	typedef std::vector<seg_index> seg_vect;

	/**  Constant seg_vect iterator data type.
	  *  \brief Segment list iterator type.
	  */ 
	typedef seg_vect::const_iterator const_seg_iter;

	/**  Enumerate all the commands to be recognized by the server.
	  *  \brief Server command enumeration.
	  */
	enum command_code {
	    kNoop,             		///< No operation
	    kGetStatus,        		///< Fetch server status
	    kGetData,          		///< Fetch channel data
	    kGetOnlineData,    		///< Fetch online channel data
	    kGetServerStatus,  		///< Fetch server status
	    kGetChannels,      		///< Fetch a list of channels
	    kCountChannels,    		///< Count channels
	    kSegmentStatus,    		///< Print segment status
	    kQuit,             		///< Cancel transaction, disconnect
	    kGetSourceList,    		///< Fetch list of sources of specified channels
	    kGetSourceData,    		///< Fetch available data sources
	    kSetEpoch,         		///< Set epoch
	    kChannelCRC,       		///< Return channel list checksum
	    kServerProtocolVersion,	///< Version of the Server Protocol
	    kServerProtocolRevision,	///< Revision of the Server Protocol
	    kListEpochs,                ///< List of epochs
	    kCheckData,                 ///< Check data availability
	    kGetLastMessage,            ///< Get the most recent error message
	    kCommandCount      		///< Number of commands.
	};

	/**  Enumerate the write response states.
	  */
	enum write_state {
	    ws_initial,    ///<  Stating state, no response has been given
	    ws_response,   ///<  Initial response (tid, flag) has been sent 
	                   ///<    to the client.
	    ws_chandata,   ///<  Channel data have been sent to the client.
	    ws_end_data    ///<  No more data available
	};

    public:
	/**  Construct a transaction block and note the queue into which 
	  *  the transaction is to be returned when processing is complete.
	  *  @brief Transaction constructor.
	  *  @param retQ Queue into which the trasaction is to be returned.
	  *  @param at   Authorization type.
	  */
	transaction(tran_queue& retQ, auth_server::auth_type at);

	/**  Destroy the transaction. Release all allocated resources.
	  */
	~transaction(void);

	/**  Return and increment reply block sequence number.
	  *  @brief get block sequence number.
	  *  @return Block sequence
	  */
	long block_seq(void);

	/**  Add a channel index to the transaction channel list.
	  *  @brief Add a channel.
	  *  @param inx Index of the channel to be processed.
	  */
	void add_channel(chan_index inx);

	/**  Authenticate the client user and get his/her user info.
	  *  @brief Get a valid user name/realm.
	  *  @param b Scratch buffer for use in authentication protocol.
	  */
	void authenticate(buffer& b);

	/**  Get a reference to the channel status structure.
	  *  @param i Index into the trasaction channel list.
	  *  @return Reference to the specified channel status structure.
	  */
	chan_stat& chan(int i);

	/**  Get a reference to the channel status structure.
	  *  @param i Index into the trasaction channel list.
	  *  @return Constant status reference for the specified channel.
	  */
	const chan_stat& chan(int i) const;

	/**  Clear transaction descriptor.
	  */
	void clear(void);

	/**  Release all segments and clear the channel list.
	  *  \brief Clear the channel list.
	  */
	void clear_channels(void);

	/**  Get command code.
	  *  @return Transaction command code.
	  */
	command_code command(void) const;

	/**  Write a formatted dump of the transaction descriptor to the 
	  *  specified output stream.
	  *  @brief Dump out the transaction.
	  *  @param out Output stream to which the dump will be sent.
	  *  @return The output stream specified by the argument.
	  */
	std::ostream& dump(std::ostream& out) const;

	/**  Get the first gps time needed by the selected channel.
	  *  @brief Get fill point.
	  *  @param i Index into transaction channel list.
	  *  @return Fill time for indicated channel.
	  */
	gps_type getChanFill(int i) const;

	/**  Get the master channel list index of the ith channel in this 
	  *  transaction.
	  *  @brief Get ith channel index
	  *  @param i Index into transaction list of channel.
	  *  @return Index in the master channel list of the requested channel.
	  */
	chan_index getChanIndex(int i) const;

	/**  Check whether all data are available in reserved segments for 
	  *  the i'th channel in this transaction.
	  *  @brief Test if channel is ready
	  *  @param i Cindex into channel list for this transaction.
	  *  @return True if all channel data are in reserved segments.
	  */
	bool getChanReady(int i) const;

	/**  Convert a command name string into a command code.
	  *  \exception \c runtime_error if the command string is not 
	  *             recognized.
	  *  \brief Get command code.
	  *  \param str Command name string.
	  *  \return Command code.
	  */
	static command_code getCommandCode(const std::string& str);

	/**  Convert a command code into a command name string.
	  *  @brief Get command string.
	  *  @param code Command code.
	  *  @return Command name string.
	  */
	static const char*  getCommandString(command_code code);

	/**  Get the default channel type code..
	  *  @brief Get the default channel type.
	  *  @return Default channel type.
	  */
	chan_type getDefType(void) const;

	/**  Get the number of channels requested for this transaction.
	  *  @brief Get number of channels.
	  *  @return Number of channels requested for this transaction.
	  */
	int getNChan(void) const;

	/**  Get the number of the last stage to have been processed.
	  *  \brief Get the last stage to have processed this transaction.
	  *  \return Stage ID.
	  */
	long get_stage(void) const;

	/**  Get the writer state or this transaction.
	  *  \brief Get the writer state
	  *  \return Writer state.
	  */
	write_state get_state(void) const;

	/**  Get the start time of this transaction.
	  *  @brief Get start time
	  *  @return Start gps
	  */
	gps_type getStartGPS(void) const;

	/**  Get the end time of this transaction.
	  *  @brief Get end time
	  *  @return Stop GPS
	  */
	gps_type getStopGPS(void) const;

	/**  Get the end time of the current stride of this transaction.
	  *  @brief Get stride end time
	  *  @return End od stride GPS
	  */
	gps_type getStrideEnd(void) const;

	/**  home queue name
	 */
	const std::string& home_queue(void) const;

	/**  Return the transaction identifier.
	  *  @brief Transaction ID.
	  *  @return Transaction identifier.
	  */
	long ident(void) const;

	/**  The stride start time is set to the end of the current stride.
	  *  All reserved segments covering the just completed stride are 
	  *  released. If there are more data to be transferrred for this 
	  *  transaction, next_stride treturns true.
	  *  @brief Set time for next stride.
	  *  @return True if transaction is not complete.
	  */
	bool next_stride(void);

	/**  Get the protocol revision to be used for this transaction.
	  *  \brief Return transaction protocol revision.
	  *  \return Transaction protocol revision.
	  */
	long protocol(void) const;
	
	/**  Sent a response code to the client int he form of a 4-digit
	  *  hex number.
	  *  \brief Respond to client.
	  *  \param resp Response code.
	  *  \return Number of bytes written.
	  */
	int put_response(int resp);

	/**  Return a reference to the epoch to which the data requests
	  *  are to be limited.
	  *  @brief Reference %epoch structure.
	  *  @return Reference to %epoch structure.
	  */
	const epoch& ref_epoch(void) const;

	/**  Return a constant reference to the latest error message 
	  *  for this transaction
 	  *  \brief Get the latest error message.
	  *  \return reference to latest error message string.
 	  */
	const std::string& ref_message(void) const;	

	/**  Return a constant reference to the channel name pattern string.
 	  *  \brief Pattern string reference.
	  *  \return Reference to pattern string.
 	  */
	const std::string& ref_pattern(void) const;	

	/**  Return a reference to the OpenSSL I/O base structure used by this
	  *  transaction.
	  *  @brief Reference OpenSSL bio structure.
	  *  @return Reference to OpenSSL bio structure.
	  */
	socket_api& ref_socket(void);

	/**  Return a reference to the user data structure for the user
	  *  authenticated for this transaction.
	  *  @brief Reference user_data structure.
	  *  @return Reference to user_data structure.
	  */
	const user_data& ref_user(void) const;

	/**  Read a line from the OpenSSL channel ito the specified buffer.
	  *  @brief read a line from the client.
	  *  @param buf Buffer into which the client data is to be read.
	  *  @return Number of characters in the line.
	  */
	int  read(buffer& buf);

	/**  Send a reply to the client.
	  *  @brief Send a reply.
	  *  @param reply Buffer containing the reply.
	  */
	void reply(const buffer& reply);

	/**  Release all transaction resources, close the connection and 
	  *  return the transaction structure to the free transaction queue.
	  *  @brief Release the transaction.
	  */
	void release(void);

	/**  Release all segments being used by this transaction.
	  *  @brief release all segments.
	  */
	void release_segments(void);

	/**  Release all of the segments being used by this transaction that
	  *  contain only data before the specified gps.
	  *  @brief Release segments with data before the specified GPS.
	  *  @param before Earliest end time for data to be released.
	  */
	void release_segments(gps_type before);

	/**  Reserve all available segments whilch will be used by this
	  *  transaction.
	  *  @brief Reserve all available segments.
	  */
	void reserve_segments(void);

	/**  Release any resources and clear command dependent fields.
	  *  @brief Reset command
	  */
	void reset(void);

	/**  Test whether the client is authenticated by the auth_server.
	  *  @brief Test if secure authentication
	  *  @return true if secure protocol used.
	  */
	bool secure(void) const;

	/**  set_auth_debug
 	  *
	  */
	void set_auth_debug(int dbg);

	/**  Set the transaction command code.
	  *  @brief Set the command.
	  *  @param cmd Transacton command code.
	  */
	void setCommand(command_code cmd);

	/**  Translate a string to a transaction command code and set the 
	  *  transaction command.
	  *  @brief Set the command from a string.
	  *  @param cmd Transacton command string.
	  */
	void setCommand(const std::string& cmd);

	/**  Set the protocol revieion to be used by default on all 
	  *  transactions.
	  *  \brief Set default protocol
	  *  \param proto Default protocol revision.
	  */
	static void set_default_protocol(long proto);
	
	/**  Set the default channel type code..
	  *  @brief Get the default channel type.
	  *  @param t Default channel type code.
	  */
	void setDefType(chan_type t);

	/**  Set the transaction ID (serial number).
	  *  @brief set the transaction id.
	  *  @param tranid Transaction serial number.
	  */
	void set_id(long tranid);

	/**  Set the transaction validity %epoch
	  *  \brief set the transaction %epoch.
	  *  \param epic Transaction %epoch
	  */
	void set_epoch(const epoch& epic);

	/**  Record the most recent error message.
	  *  \brief Record error message text.
	  *  \param estage current stage.
	  *  \param what error message text
	  */
	void set_message(const std::string& estage, const std::string& what);
	
	/**  Set the pattern to be used in a pattern-matching channel list 
 	  *  request.
 	  *  \brief Set channel name pattern
 	  *  \param pat Pattern string
 	  */
	void set_pattern(const std::string& pat);

	/**  Set the protocol revision to be used by this transaction.
	  *  \brief Set default protocol
	  *  \param proto Defult protocol number.
	  */
	void set_protocol(long proto);

	/**  Identify the communication socket to be used by this transaction.
	  *  @brief Set socket for this transaction.
	  *  @param sock Pointer to a socket_api object.
	  */
	void set_socket(socket_api* sock);

	/**  Set the ID of the transaction stage being executed. This may help
	  *  in debugging.
	  *  @brief Set transaction stage.
	  *  @param s Identity of the last stage executed.
	  */
	void set_stage(long s);

	/**  Set the writer state.
	  *  @brief Set writer state
	  *  @param s New write state code.
	  */
	void set_state(write_state s);

	/**  Set the transaction time range and stride. If a stride is 
	  *  specified, the transaction will be divided into a number 
	  *  of replies with the specified length. If the stride is 
	  *  unspecified or zero, a single reply is returned.
	  *  @brief Set transaction times.
	  *  @param start  Start GPS time.
	  *  @param stop   End GPS time.
	  *  @param stride Time stride in seconds.
	  */
	void setTime(gps_type start, gps_type stop, gps_type stride=0);

	/**  Set the channel fill times to the start time of this transaction.
	  *  @brief Set channel fill status to start time.
	  *  @param start  Start GPS time.
	  */
	void start_fill(gps_type start);

	/**  Get the stride length
	  *  \brief Time stride;
	  *  \return time stride in seconds.
	  */
	gps_type stride(void) const;

	/**  Test whether the connection to the client is still open (\e i.e. 
	  *  that the client has not hung up.
	  *  \brief Test client connection
	  *  \return True if te connection is open
	  */
	bool test_connect(void) const;
	
	/**  Test the hashed index against the stored crc. If the crc differs
	  *  the test returns true and the crc is updated.
	  *  \brief test for a different index block.
	  *  \param hash String representating the index contents
	  *  \return Hash has changed from its previous value.
	  */
	bool test_index_crc(const std::string& hash);

	/**  Estimate the buffer resources needed for this transaction and
	  *  reject the transaction (false return) if the requested data is 
	  *  excessive. At present the usage of each class of buffers is 
	  *  estimated. The request is rejected if the requirements for any 
	  *  class of buffers excedes 80% of the total number of segment 
	  *  buffers in the class.
 	  *  @brief Test requested resource usage
	  *  @return True if the resource usage is reasonable.
	  */
	bool test_limits(void) const;

	/**  Write state string.
	 */
	static std::string write_state_string(write_state s);

    private:
	static long default_protocol;
	long trans_id;
	long trans_protocol;
	tran_queue* home;
	socket_api* trans_sock;
	auth_server trans_auth;
	command_code trans_code;
	gps_type trans_start;
	gps_type trans_stop;
	gps_type trans_stride;
	cinx_vect trans_channels;
	seg_vect trans_segments;
	write_state trans_state;
	uint32_t trans_index_crc;
	long trans_blockid;
	chan_type trans_chant;
	long stage;
	epoch trans_epoch;
	std::string trans_pattern;
	std::string last_message;
    };


    /**  Add a name field to the tran_queue.
      */
    class tran_queue : public thread::mt_queue<transaction*>
    {
    public:
	tran_queue(int nqueue, const std::string& qname)
	    : thread::mt_queue<transaction*>(nqueue), _name(qname) {}
	const std::string& name(void) const {return _name;}
	void set_name(const std::string& qname) {_name = qname;}
    private:
	std::string _name;
    };

    //==================================  transaction::chan_stat inline methods
    inline 
    transaction::chan_stat::chan_stat(chan_index chn)
	: _chinx(chn), _fillpt(0), _status(0), _driver(0)
    {}

    inline 
    transaction::chan_stat::chan_stat(void) {
    }

    inline bool
    transaction::chan_stat::available(void) const {
	return !_status;
    }

    inline driver_api* 
    transaction::chan_stat::driver(void) const {
	return _driver;
    }

    inline gps_type
    transaction::chan_stat::fill(void) const {
	return _fillpt;
    }

    inline chan_index 
    transaction::chan_stat::index(void) const {
	return _chinx;
    }

    inline bool 
    transaction::chan_stat::ready(gps_type t) const {
	return _fillpt >= t || !available();
    }

    inline void 
    transaction::chan_stat::set_driver(driver_api* d) {
	_driver = d;
    }

    inline void 
    transaction::chan_stat::set_fill(gps_type t) {
	_fillpt = t;
    }

    inline void 
    transaction::chan_stat::set_stat(int e) {
	_status = e;
    }

    inline int 
    transaction::chan_stat::status(void) const {
	return _status;
    }

    //==================================  transction inline methods
    inline long 
    transaction::block_seq(void) {
	return trans_blockid++;
    }

    inline transaction::chan_stat&
    transaction::chan(int i) {
	return trans_channels[i];
    }

    inline const transaction::chan_stat&
    transaction::chan(int i) const {
	return trans_channels[i];
    }

    inline transaction::command_code 
    transaction::command(void) const {
	return trans_code;
    }

    inline gps_type
    transaction::getChanFill(int i) const {
	gps_type t = trans_channels[i].fill();
	if (t < trans_start) t = trans_start;
	return t;
    }

    inline chan_index
    transaction::getChanIndex(int i) const {
	return trans_channels[i].index();
    }

    inline bool
    transaction::getChanReady(int i) const {
	return trans_channels[i].ready(getStrideEnd());
    }

    inline chan_type
    transaction::getDefType(void) const {
	return trans_chant;
    }

    inline int
    transaction::getNChan(void) const {
	return trans_channels.size();
    }

    inline long
    transaction::get_stage(void) const {
	return stage;
    }

    inline transaction::write_state
    transaction::get_state(void) const {
	return trans_state;
    }

    inline gps_type 
    transaction::getStartGPS(void) const {
	return trans_start;
    }

    inline gps_type  
    transaction::getStopGPS(void) const {
	return trans_stop;
    }

    inline gps_type  
    transaction::getStrideEnd(void) const {
	return trans_start + stride();
    }

    inline long
    transaction::ident(void) const {
	return trans_id;
    }

    inline long 
    transaction::protocol(void) const {
	return trans_protocol;
    }

    inline const epoch& 
    transaction::ref_epoch(void) const {
	return trans_epoch;
    }

    inline const std::string& 
    transaction::ref_message(void) const {
	return last_message;
    }

    inline const std::string& 
    transaction::ref_pattern(void) const {
	return trans_pattern;
    }

    inline socket_api&
    transaction::ref_socket(void) {
	return *trans_sock;
    }

    inline const user_data& 
    transaction::ref_user(void) const {
	return trans_auth.refUser();
    }

    inline bool 
    transaction::secure(void) const {
	return trans_auth.secure();
    }

    inline gps_type 
    transaction::stride(void) const {
	if (trans_stop && trans_stop - trans_start < trans_stride) {
	    return trans_stop - trans_start;
	}
	return trans_stride;
    }

}

#endif // !defined(SENDS_TRANSACTION_HH)
