/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_UPDATE_LIST_HH
#define SENDS_UPDATE_LIST_HH

#include "thread_base.hh"

#include <vector>
#include <string>


namespace sends {

    class src_update;

    /**  The update_list class maintains a list of source update requests
      *  and provides global update thread launching and termination.
      *  \brief List of source update requests.
      *  \author John Zweizig
      *  \version 2009-05-15
      *  \ingroup sends_base
      */
    class update_list {
    public:
	/**  Vector of src_update request objects.
	  */
	typedef std::vector<src_update*>    update_vect;
	/**  Iterator over source update requests
	  */
	typedef update_vect::iterator       update_iterator;
	/**  Constant iterator over source update requests
	  */
	typedef update_vect::const_iterator const_update_iterator;


	/**  Vector of threaded requests
	  */
	typedef std::vector<std::pair<std::string, thread::thread_base*> > thread_vect;
	/**  Iterator over general update threads
	  */
	typedef thread_vect::iterator 		thread_iterator;
	/**  Constant iterator over a thread 
	  */
	typedef thread_vect::const_iterator const_thread_iterator;

    public:
	/** Update list constructor.
	  */
	update_list(void);

	/** Update list destructor.
	  */
	~update_list(void);

	/**  Add an update request to the list.
	  *  \param su constant reference to the src_update.
	  */
	void add(const src_update& su);

	void add(const std::string& name, thread::thread_base* gen_thread);

	/**  Start all update threads.
	  */
	void start_all(void);

	/**  Stop all update threads in the list.
	  */
	void stop_all(void);

	/**  Find the source updater matching the specified name.
	  *  \param name Name of the updater to be found
	  *  \return Pointer to the src_update class object.
	  */
	src_update* find(const std::string& name);

	thread::thread_base* find_thread(const std::string& name);

	/**  Find the source updater matching the specified name.
	  *  \param name Name of the updater to be found
	  *  \return Constant pointer to the src_update class object.
	  */
	const src_update* find(const std::string& name) const;

	const thread::thread_base* find_thread(const std::string& name) const;

	/** test whether the specified type has an updater defined.
	 */
	bool update_type(const std::string& name) const;

    private:
	/**  Vector of src_update thread object pointers.
	  */
	update_vect mVect;
	thread_vect mThreads;
    };

    extern update_list UpdateList;
}

#endif  //  !defined(SENDS_UPDATE_LIST_HH)
