/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_USER_DATA_HH
#define SENDS_USER_DATA_HH
#include "types.hh"
#include "gmutex.hh"
#include <string>
#include <map>
#include <ctime>

namespace sends {

    /**  The user_data class contains information about users and their 
      *  acess to channel data.
      *  @brief User authorization data.
      *  @author John Zweizig
      *  @version 1.0; Las modified January 23, 2008
      *  @ingroup sends_base
      */
    class user_data {
    public:

	/**  Default user data constructor.
	  *  @brief default constructor.
	  */
	user_data(void) {}

	/**  Construct a user_data object and fill it with user information.
	  *  @brief Data constructor.
	  *  @param name User name.
	  *  @param pwd  User's encoded password.
	  *  @param access User's access level.
	  */
	user_data(const std::string& name, const std::string& pwd,
		   long access=0);

	/**  Erase and destroy a user_data object.
	  *  @brief Destructor.
	  */
	~user_data(void);

	/**  Erase the user data.
	  *  @brief Erase all user data.
	  */
	void erase(void);

	/**  Return a constant reference to a string containing the user name.
	  *  @brief User name.
	  *  @return Constant reference to a string containing the user name.
	  */
	const std::string& user(void) const;

	/**  Return a constant reference to a string containing the 
	  *  encrypted password.
	  *  @brief Encrypted password.
	  *  @return Constant reference to a string with the encrypted password.
	  */
	const std::string& passwd(void) const;

	/**  Return the user's channel access level.
	  *  @brief Access level.
	  *  @return the Users access level.
	  */
	access_mask access(void) const;

    private:
	std::string mUserName;
	std::string mKeyWord;
	access_mask mAccess;
    };

    //==================================  Inline user_data methods.
    inline const std::string& 
    user_data::user(void) const {
	return mUserName;
    }

    inline const std::string& 
    user_data::passwd(void) const {
	return mKeyWord;
    }

    inline access_mask
    user_data::access(void) const {
	return mAccess;
    }

    /**  The user_list class is a list of user_data acceible by the user name
      *  field.
      *  @brief List of known users.
      *  @author John Zweizig
      *  @version 1.0; Last Modified January 23, 2008
      *  @ingroup sends_base
      */
    class user_list {
    public:
	/**  User list data type keyed by the user name.
	  */
	typedef std::map<std::string, user_data> user_map;

	/**  User list iterator
	  */
	typedef user_map::iterator user_iterator;

	/**  Constant user list iterator
	  */
	typedef user_map::const_iterator const_user_iterator;
    public:
	/**  Construct an empty user list.
	  *  @brief Empty list constructor.
	  */
	user_list(void);

	/**  Destroy a user list.
	  *  @brief Destructor.
	  */
	~user_list(void);

	/**  Dump out a list of users and their access.
	  *  @brief Dump the list.
	  *  @param ostr output stream
	  */
	std::ostream& dumpUsers(std::ostream& ostr) const;

	/**  Test whether the specified user is known.
	  *  @brief Test if a user us known
	  *  @param user Name of user
	  *  @return true If user known
	  */
	bool known(const std::string& user) const;

	/**  Fill a list from a specified ascii file.
	  *  @brief Fill the list.
	  *  @param file Pat to an Ascii file containing the user information.
	  */
	void readFile(const std::string& file);

	/**  Return a constant reference to the user_data object for the 
	  *  specified user name.
	  *  @brief Get user data.
	  *  @param name Constant reference to a string with the user name.
	  *  @exception runtime_error if the named user is not defined.
	  *  @return Constant reference to the user data object.
	  */
	const user_data& operator[](const std::string& name) const;

	/** Return the timestamp of the userfile that was loaded.
	  * @return std::time_t timestamp value of the file mtime when it was
	  * read in.  0 if no file has been read.
	  */
	std::time_t timestamp() const;

	/** Swap the contents of two user_lists objects.
	  * @param other the object to swap contents with.
	  */
	void swap(user_list& other);
    private:
	user_map mUserMap;
	std::time_t mTimestamp;
	mutable thread::readwritelock mRWLock;
    };

    //==================================  Define external user list.
    extern user_list User_List;
}

#endif // SENDS_USER_DATA_HH
