/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_WRITE_STAGE_HH
#define SENDS_WRITE_STAGE_HH

#include "seda_stage.hh"

namespace sends {

    class chan_entry;
    class epoch;

    /**  The write_stage class provides a minimal template for a SEDA 
      *  stage thread pool. It is based on the seda_stage class and provides
      *  a stage with no filter processing. The constructor, destructor and 
      *  filter methods are implemented. The filter method immediately passes 
      *  the transactions to its output or alternate queue (the input queue 
      *  of the next stage in the pipeline. In the case of the write stage, 
      *  the standard output is the parse stage (execute another command) and 
      *  the alternate output is the fetch stage (fetch more data for this 
      *  transaction).
      *  @brief Data reply SEDA stage.
      *  @author J. Zweizig
      *  @version 1.1; Last modified May 13, 2009
      *  @ingroup sends_stage
      */
    class write_stage : public seda_stage {
    public:
	/**  Construct a write_stage object. The arguments indicate the 
	  *  number of threads in the pool and the number of input transactions
	  *  that may be queue simultaneously (note that the pipline gets 
	  *  indigestion only after more than (nTrans + nThread) transactions
	  *  need to execute in this stage.
	  *  @brief Stage template constructor.
	  *  @param nThread Number of threads to run simultaneously in the pool.
	  *  @param nTrans  Number of transactions that may be queued.
	  */
	write_stage(int nThread, int nTrans);

	/**  Stage template destructor is null by default. It is mostly
	  *  useful to remove resources allocated by the stages.
	  */
	virtual ~write_stage(void);

	/**  The filter method performs the transaction processing for a 
	  *  single stage. By default this is a null filter that passes 
	  *  the transaction on to the next 
	  *  @brief Transaction filter for this stage.
	  *  @param trans Pointer to the current transaction 
	  */
	void filter(transaction* trans);

	/**  Set up the signal handling for this thread.
	  *  \brief Thread initialization.
	  */
	void init_thread(void);

	/**  Return a constant pointer to a null-terminated character vector 
	  *  containing the name of the stage.
	  *  @brief Stage name.
	  *  @return Constant pointer to the stage name string.
	  */
	const char* stage_name(void) const;

    private:
	///  Index vector.
	typedef std::vector<chan_index> index_vect;

	/**  Build a channel index satisfying the transaction requirements.
	  *  The channel index vector is filled with the channel entry number
	  *  of each channel satisfying the transaction requirements (gps and
	  *  channel type). If all channel list entries satify the requirements
	  *  the channel list may be empty.
	  *  \brief Build a channel index list
	  *  \param trans    Constant reference to a transaction instance.
	  *  \param chanlist Channel index vector.
	  *  \return Number of channels in the list
	  */
	chan_index build_channel_index(const transaction& trans, 
				       index_vect& chanlist) const;

	/**  Test whether the specified channel (index) matches the channel 
	  *  type and is defined at the gps time. If the specified
	  *  channel type is kUnknown or the GPS time is 0, the corresponding
	  *  test is omitted. If gps is non-zero, only persisitent (not online)
	  *  channels qualify and the time must be in a defiend source epoch.
	  *  \brief Test whether channel belongs in specified sub-list.
	  *  \param c  Channel type the channel must match (or kUnknown).
	  *  \param t  GPS time the channel must be available (or 0).
	  *  \param ce Channel entry to be tested
	  *  \return True if the channel matches the type and time criteria.
	  */
	bool test_channel(chan_type c, gps_type t, 
			  const chan_entry& ce) const;

	/**  Test whether the specified channel (index) matches the channel 
	  *  type and is defined at the gps time. If the specified
	  *  channel type is kUnknown or the GPS time is 0, the corresponding
	  *  test is omitted.
	  *  \brief Test whether channel belongs in specified sub-list.
	  *  \param c    Channel type the channel must match (or kUnknown).
	  *  \param epic Epoch the channel must be available (or 0).
	  *  \param ce   Channel entry to be tested
	  *  \return True if the channel matches the type and time criteria.
	  */
	bool test_channel(chan_type c, const epoch& epic, 
			  const chan_entry& ce) const;

	/**  Write channel status block defining all requested channels.
	  *  \brief Write transaction status block.
	  *  \param trans Pointer to current transaction description.
	  */
	void write_status(transaction* trans);

	/**  Write channel data block with data for the current stride.
	  *  \brief Write transaction channel data.
	  *  \param trans Pointer to current transaction description.
	  */
	void write_data(transaction* trans);

	/**  Write a list of defined epochs in reply to the list-epochs command
	  *  \brief Write a list of epoch definitions
	  *  \param trans Pointer to current transaction description.
	  */
	void write_epochs(transaction* trans);

	/**  Write a server status response block.
	  *  \brief Write server status block.
	  *  \param trans Pointer to current transaction description.
	  */
	void write_server_status(transaction* trans);

	/**  Write a source list to the client socket. The source list includes
	  *  all sources in which the specified channel(s) are defined and a
	  *  list of time epochs during which they are defined in the source.
	  *  If a start time is specified in the client request the time spans
	  *  are omitted and only the sources with data available at the 
	  *  specified time are added to the list. The \a avail argument 
	  *  causes the defined data times to me modified to reflect frames 
	  *  that are currently available to the server.
	  *  \brief Write a source list.
	  *  \param trans Pointer to current transaction description.
	  *  \param avail If true, times reflect currently avalable data.
	  */
	void write_source_list(transaction* trans, bool avail);

	/**  Write a channel list response block.
	  *  \brief Write a channel list response block.
	  *  \param trans Pointer to current transaction description.
	  */
	void write_channel_list(transaction* trans);

	/**  Write a channel count response.
	  *  \brief Write a channel count response.
	  *  \param trans Pointer to current transaction description.
	  */
	void write_channel_count(transaction* trans);

	/**  Write a channel list crc response.
	  *  \brief Write a channel list crc response.
	  *  \param trans Pointer to current transaction description.
	  */
	void write_channel_crc(transaction* trans);

    private:
    };
}

#endif // !defined(SENDS_WRITE_STAGE_HH)

